time: float = 0.0
k1: float = 0.900000000000000
p1: float = 0.100000000000000
p2: float = 1.50000000000000
S1: float = 1.50000000000000
S2: float = 0.0

# Initial assignments
C = p1 * p2
S1_conc = S1 / C
S2_conc = S2 / C
dp2 = 0.100000000000000
reaction1 = S1_conc * k1
y0 = [S1, S2, p2]
variable_names = ["S1", "S2", "p2"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, S2, p2 = variables
    C: float = p1 * p2
    S1_conc: float = S1 / C
    dp2: float = 0.100000000000000
    reaction1: float = S1_conc * k1
    dp2dt: float = dp2
    dS1dt: float = -C * reaction1
    dS2dt: float = C * reaction1
    return dS1dt, dS2dt, dp2dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, S2, p2 = variables
    C: float = p1 * p2
    S1_conc: float = S1 / C
    S2_conc: float = S2 / C
    dp2: float = 0.100000000000000
    reaction1: float = S1_conc * k1
    return {
        "C": C,
        "S1_conc": S1_conc,
        "S2_conc": S2_conc,
        "dp2": dp2,
        "reaction1": reaction1,
    }



time: float = 0.0
p1: float = 5.00000000000000

# Initial assignments
p2 = time + 1
y0 = []
variable_names = []


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    return


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    p2: float = time + 1
    return {
        "p2": p2,
    }



time: float = 0.0
C: float = 1.00000000000000
S1_conc: float = 3.00000000000000
S2: float = 3.00000000000000

# Initial assignments
S1_stoich = 2
S1 = C * S1_conc
J0 = 0.1 * S1
y0 = [S2]
variable_names = ["S2"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    (S2,) = variables
    S1_stoich: float = 2
    S1: float = C * S1_conc
    J0: float = 0.1 * S1
    dS2dt: float = J0
    dS1dt: float = -J0 * S1_stoich
    return (dS2dt,)


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    (S2,) = variables
    S1_stoich: float = 2
    S1: float = C * S1_conc
    J0: float = 0.1 * S1
    return {
        "S1_stoich": S1_stoich,
        "S1": S1,
        "J0": J0,
    }



time: float = 0.0
p1: float = 0.0

# Initial assignments
J0 = 3
dp1 = J0
y0 = [p1]
variable_names = ["p1"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    (p1,) = variables
    J0: float = 3
    dp1: float = J0
    dp1dt: float = dp1
    return (dp1dt,)


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    (p1,) = variables
    J0: float = 3
    dp1: float = J0
    return {
        "J0": J0,
        "dp1": dp1,
    }



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, cast

import libsbml

from . import data, l1

if TYPE_CHECKING:
    from .data import Model

__all__ = [
    "data",
    "parse",
]


def parse(
    lib_model: libsbml.Model,
    level: int,  # Literal[1, 2, 3],
    version: int,  # Literal[1, 2, 3, 4, 5],
) -> Model:
    """Parse sbml model."""

    # FIXME: actually use different parsers here
    match version:
        case 1:
            return l1.parse(lib_model=lib_model, level=level)
        case 2:
            return l1.parse(lib_model=lib_model, level=level)
        case 3:
            return l1.parse(lib_model=lib_model, level=level)
        case _:
            msg = f"Unknown SBML version {version}"
            raise NotImplementedError(msg)


def load_document(file: str | Path) -> data.Document:
    if not Path(file).exists():
        msg = "Model file does not exist"
        raise OSError(msg)

    doc = libsbml.readSBMLFromFile(str(file))
    version = cast(int, doc.version)
    level = cast(int, doc.level)

    model: libsbml.Model = doc.getModel()
    if len(model.getListOfAllElementsFromPlugins()):
        msg = "Plugin handling not yet implemented"
        raise NotImplementedError(msg)

    return data.Document(
        model=parse(model, version=version, level=level),
        plugins=[
            data.Plugin(name=(doc.getPlugin(i).package_name))
            for i in range(doc.num_plugins)
        ],
    )



time: float = 0.0

# Initial assignments
a = 2
b = a + 1
c = b + 1
d = c + 1
e = d + 1
f = e + 1
y0 = []
variable_names = []


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    return


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    b: float = a + 1
    d: float = c + 1
    f: float = e + 1
    return {
        "b": b,
        "d": d,
        "f": f,
    }



time: float = 0.0
k3: float = 1.00000000000000
k1: float = 0.000150000000000000
k2: float = 0.0

# Initial assignments
dk1 = -k1 * k3 * time
dk2 = k1 * k3 * time
y0 = [k1, k2]
variable_names = ["k1", "k2"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    k1, k2 = variables
    dk1: float = -k1 * k3 * time
    dk2: float = k1 * k3 * time
    dk1dt: float = dk1
    dk2dt: float = dk2
    return dk1dt, dk2dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    k1, k2 = variables
    dk1: float = -k1 * k3 * time
    dk2: float = k1 * k3 * time
    return {
        "dk1": dk1,
        "dk2": dk2,
    }



time: float = 0.0

# Initial assignments
P1 = 6.02214179000000e23
P2 = 6.02214179000000e23
P3 = 6.02214179000000e23
P4 = 6.02214179000000e23
P5 = 4.17899999992689e18
y0 = []
variable_names = []


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    return


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    return {}



time: float = 0.0
k1: float = 0.100000000000000
k2: float = 0.150000000000000
p1: float = 2.50000000000000
C: float = 2.50000000000000
S1: float = 1.00000000000000
S2: float = 0.0
S3: float = 0.0

# Initial assignments
S4 = S3 / (p1 + 1)
S5 = S4 * p1
reaction1 = S1 * k1
reaction2 = S5 * k2
y0 = [S1, S2, S3]
variable_names = ["S1", "S2", "S3"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, S2, S3 = variables
    S4: float = S3 / (p1 + 1)
    S5: float = S4 * p1
    reaction1: float = S1 * k1
    reaction2: float = S5 * k2
    dS1dt: float = -reaction1
    dS3dt: float = reaction1 - reaction2
    dS2dt: float = reaction2
    return dS1dt, dS2dt, dS3dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, S2, S3 = variables
    S4: float = S3 / (p1 + 1)
    S5: float = S4 * p1
    reaction1: float = S1 * k1
    reaction2: float = S5 * k2
    return {
        "S4": S4,
        "S5": S5,
        "reaction1": reaction1,
        "reaction2": reaction2,
    }



time: float = 0.0
k1: float = 1.00000000000000
default_compartment: float = 1.00000000000000
p1: float = 1.00000000000000
X: float = 1.00000000000000

# Initial assignments
Xref = p1
dp1 = 1
J0 = k1
y0 = [X, p1]
variable_names = ["X", "p1"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    X, p1 = variables
    Xref: float = p1
    dp1: float = 1
    J0: float = k1
    dp1dt: float = dp1
    dXdt: float = J0 * Xref
    return dXdt, dp1dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    X, p1 = variables
    Xref: float = p1
    dp1: float = 1
    J0: float = k1
    return {
        "Xref": Xref,
        "dp1": dp1,
        "J0": J0,
    }



time: float = 0.0
k2: float = 300000.000000000
compartment: float = 1.00000000000000
k1: float = 1000000.00000000
S1: float = 1.00000000000000e-6
S2: float = 1.50000000000000e-6
S3: float = 2.00000000000000e-6
S4: float = 5.00000000000000e-7

# Initial assignments
S1_conc = S1 / compartment
S2_conc = S2 / compartment
S3_conc = S3 / compartment
S4_conc = S4 / compartment
dk1 = 1000000
reaction1 = S1_conc * S2_conc * k1
reaction2 = S3_conc * S4_conc * k2
y0 = [S1, S2, S3, S4, k1]
variable_names = ["S1", "S2", "S3", "S4", "k1"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, S2, S3, S4, k1 = variables
    S1_conc: float = S1 / compartment
    S2_conc: float = S2 / compartment
    S3_conc: float = S3 / compartment
    S4_conc: float = S4 / compartment
    dk1: float = 1000000
    reaction1: float = S1_conc * S2_conc * k1
    reaction2: float = S3_conc * S4_conc * k2
    dk1dt: float = dk1
    dS1dt: float = -compartment * reaction1 + compartment * reaction2
    dS2dt: float = -compartment * reaction1 + compartment * reaction2
    dS3dt: float = compartment * reaction1 - compartment * reaction2
    dS4dt: float = compartment * reaction1 - compartment * reaction2
    return dS1dt, dS2dt, dS3dt, dS4dt, dk1dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, S2, S3, S4, k1 = variables
    S1_conc: float = S1 / compartment
    S2_conc: float = S2 / compartment
    S3_conc: float = S3 / compartment
    S4_conc: float = S4 / compartment
    dk1: float = 1000000
    reaction1: float = S1_conc * S2_conc * k1
    reaction2: float = S3_conc * S4_conc * k2
    return {
        "S1_conc": S1_conc,
        "S2_conc": S2_conc,
        "S3_conc": S3_conc,
        "S4_conc": S4_conc,
        "dk1": dk1,
        "reaction1": reaction1,
        "reaction2": reaction2,
    }



time: float = 0.0
S1: float = 1.20000000000000
S1: float = 1.20000000000000

# Initial assignments
c = time + 2
S1_conc = S1 / c
dS1 = 1
y0 = [S1]
variable_names = ["S1"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    (S1,) = variables
    c: float = time + 2
    dS1: float = 1
    dS1dt: float = dS1
    return (dS1dt,)


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    (S1,) = variables
    c: float = time + 2
    S1_conc: float = S1 / c
    dS1: float = 1
    return {
        "c": c,
        "S1_conc": S1_conc,
        "dS1": dS1,
    }



import math

import scipy.special

time: float = 0.0
p1: float = 4.00000000000000
p2: float = 25.0000000000000
S1: float = 1.00000000000000
S2: float = 0.0

# Initial assignments
dS1 = -scipy.special.factorial(math.ceil(S1 * p1)) / p2
dS2 = scipy.special.factorial(math.ceil(S1 * p1)) / p2
y0 = [S1, S2]
variable_names = ["S1", "S2"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, S2 = variables
    dS1: float = -scipy.special.factorial(math.ceil(S1 * p1)) / p2
    dS2: float = scipy.special.factorial(math.ceil(S1 * p1)) / p2
    dS1dt: float = dS1
    dS2dt: float = dS2
    return dS1dt, dS2dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, S2 = variables
    dS1: float = -scipy.special.factorial(math.ceil(S1 * p1)) / p2
    dS2: float = scipy.special.factorial(math.ceil(S1 * p1)) / p2
    return {
        "dS1": dS1,
        "dS2": dS2,
    }



time: float = 0.0
k1: float = 7500.00000000000
k2: float = 2500.00000000000
C: float = 1.00000000000000
S1: float = 0.000100000000000000
S2: float = 0.000100000000000000
S3: float = 0.000200000000000000
S4: float = 0.000100000000000000

# Initial assignments
reaction1 = S1 * S2 * k1
reaction2 = S3 * S4 * k2
y0 = [S1, S2, S3, S4]
variable_names = ["S1", "S2", "S3", "S4"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, S2, S3, S4 = variables
    reaction1: float = S1 * S2 * k1
    reaction2: float = S3 * S4 * k2
    dS1dt: float = -reaction1 + reaction2
    dS2dt: float = -reaction1 + reaction2
    dS3dt: float = reaction1 - reaction2
    dS4dt: float = 2.0 * reaction1 - reaction2
    return dS1dt, dS2dt, dS3dt, dS4dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, S2, S3, S4 = variables
    reaction1: float = S1 * S2 * k1
    reaction2: float = S3 * S4 * k2
    return {
        "reaction1": reaction1,
        "reaction2": reaction2,
    }



time: float = 0.0
k1: float = 1.00000000000000
default_compartment: float = 1.00000000000000
X: float = 0.0

# Initial assignments
Xref = 3
J0 = k1
y0 = [X]
variable_names = ["X"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    (X,) = variables
    Xref: float = 3
    J0: float = k1
    dXdt: float = J0 * Xref
    return (dXdt,)


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    (X,) = variables
    Xref: float = 3
    J0: float = k1
    return {
        "Xref": Xref,
        "J0": J0,
    }



time: float = 0.0
k1: float = 0.750000000000000
k2: float = 0.250000000000000
C: float = 1.00000000000000
S1_conc: float = 0.000100000000000000
S2_conc: float = 0.000200000000000000
S3_conc: float = 0.000100000000000000

# Initial assignments
S1 = C * S1_conc
S2 = C * S2_conc
S3 = C * S3_conc
reaction1 = C * S1 * S2 * k1
reaction2 = C * S3 * k2
y0 = []
variable_names = []


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1: float = C * S1_conc
    S2: float = C * S2_conc
    S3: float = C * S3_conc
    return


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1: float = C * S1_conc
    S2: float = C * S2_conc
    S3: float = C * S3_conc
    reaction1: float = C * S1 * S2 * k1
    reaction2: float = C * S3 * k2
    return {
        "S1": S1,
        "S2": S2,
        "S3": S3,
        "reaction1": reaction1,
        "reaction2": reaction2,
    }



import math

time: float = 0.0

# Initial assignments
P = math.inf
Q = float("-inf")
R = math.nan
y0 = []
variable_names = []


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    return


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    return {}



time: float = 0.0
C: float = 1.00000000000000
A: float = 1.00000000000000
B: float = 1.00000000000000

# Initial assignments
J0 = -1
A_sr1 = 2
A_sr2 = 1
B_sr1 = 2
B_sr2 = 1
y0 = [A, B]
variable_names = ["A", "B"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    A, B = variables
    J0: float = -1
    dAdt: float = J0 * (-A_sr1 + A_sr2)
    dBdt: float = J0 * (-B_sr1 + B_sr2)
    return dAdt, dBdt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    A, B = variables
    J0: float = -1
    return {
        "J0": J0,
    }



time: float = 0.0

# Initial assignments
S1 = 7
y0 = []
variable_names = []


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    return


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1: float = 7
    return {
        "S1": S1,
    }



time: float = 0.0
y: float = 3.00000000000000
C: float = 1.00000000000000
S1: float = 1.00000000000000

# Initial assignments
S1_conc = S1 / C
k1 = y
__J0 = k1
y0 = [S1]
variable_names = ["S1"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    (S1,) = variables
    __J0: float = k1
    dS1dt: float = C * __J0
    return (dS1dt,)


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    (S1,) = variables
    S1_conc: float = S1 / C
    __J0: float = k1
    return {
        "S1_conc": S1_conc,
        "__J0": __J0,
    }



time: float = 0.0
x: float = 0.0

# Initial assignments
z = (1) if (x >= 0.49) else (0)
dx = 1
y0 = [x]
variable_names = ["x"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    (x,) = variables
    dx: float = 1
    dxdt: float = dx
    return (dxdt,)


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    (x,) = variables
    z: float = (1) if (x >= 0.49) else (0)
    dx: float = 1
    return {
        "z": z,
        "dx": dx,
    }



time: float = 0.0
k1: float = 1.00000000000000
default_compartment: float = 1.00000000000000
Y: float = 0.0
Xref: float = 1.00000000000000
X: float = 0.0

# Initial assignments
Z = Xref
dY = Xref
J0 = k1
Q = Xref
y0 = [X, Y]
variable_names = ["X", "Y"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    X, Y = variables
    dY: float = Xref
    J0: float = k1
    dYdt: float = dY
    dXdt: float = J0 * Xref
    return dXdt, dYdt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    X, Y = variables
    Z: float = Xref
    dY: float = Xref
    J0: float = k1
    return {
        "Z": Z,
        "dY": dY,
        "J0": J0,
    }



time: float = 0.0
C: float = 1.00000000000000
A: float = 1.00000000000000

# Initial assignments
J0 = -1
y0 = [A]
variable_names = ["A"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    (A,) = variables
    J0: float = -1
    dAdt: float = -3.0 * J0
    return (dAdt,)


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    (A,) = variables
    J0: float = -1
    return {
        "J0": J0,
    }



time: float = 0.0
compartment: float = 1.00000000000000
S1: float = 0.0

# Initial assignments
S1_conc = S1 / compartment
dS1 = 7
y0 = [S1]
variable_names = ["S1"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    (S1,) = variables
    dS1: float = 7
    dS1dt: float = compartment * dS1
    return (dS1dt,)


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    (S1,) = variables
    S1_conc: float = S1 / compartment
    dS1: float = 7
    return {
        "S1_conc": S1_conc,
        "dS1": dS1,
    }



import math

time: float = 0.0
k1: float = 0.750000000000000
k2: float = 0.250000000000000
p1: float = 1.25000000000000e-5
C: float = math.nan
S2: float = 0.000200000000000000
S3: float = 0.000100000000000000

# Initial assignments
reaction2 = S3 * k2
S1 = 2 * p1
reaction1 = S1 * S2 * k1
y0 = [S1, S2, S3]
variable_names = ["S1", "S2", "S3"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, S2, S3 = variables
    reaction1: float = S1 * S2 * k1
    reaction2: float = S3 * k2
    dS1dt: float = -reaction1 + reaction2
    dS2dt: float = -reaction1 + reaction2
    dS3dt: float = reaction1 - reaction2
    return dS1dt, dS2dt, dS3dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, S2, S3 = variables
    reaction1: float = S1 * S2 * k1
    reaction2: float = S3 * k2
    return {
        "reaction1": reaction1,
        "reaction2": reaction2,
    }



time: float = 0.0

# Initial assignments
a = 2
b = 4
c = 1
d = 1
e = 0
y0 = []
variable_names = []


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    return


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    a: float = 2
    b: float = 4
    c: float = 1
    d: float = 1
    e: float = 0
    return {
        "a": a,
        "b": b,
        "c": c,
        "d": d,
        "e": e,
    }



time: float = 0.0
C: float = 1.00000000000000
A: float = 50.0000000000000
B: float = 1.00000000000000

# Initial assignments
dA_sr1 = -1
dA_sr2 = 1
dB_sr1 = -1
dB_sr2 = 1
J0 = 1
A_sr1 = 2
A_sr2 = 1
B_sr1 = 2
B_sr2 = 1
y0 = [A, A_sr1, A_sr2, B, B_sr1, B_sr2]
variable_names = ["A", "A_sr1", "A_sr2", "B", "B_sr1", "B_sr2"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    A, A_sr1, A_sr2, B, B_sr1, B_sr2 = variables
    dA_sr1: float = -1
    dA_sr2: float = 1
    dB_sr1: float = -1
    dB_sr2: float = 1
    J0: float = 1
    dA_sr1dt: float = dA_sr1
    dA_sr2dt: float = dA_sr2
    dB_sr1dt: float = dB_sr1
    dB_sr2dt: float = dB_sr2
    dAdt: float = J0 * (-A_sr1 - A_sr2)
    dBdt: float = J0 * (B_sr1 + B_sr2)
    return dAdt, dA_sr1dt, dA_sr2dt, dBdt, dB_sr1dt, dB_sr2dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    A, A_sr1, A_sr2, B, B_sr1, B_sr2 = variables
    dA_sr1: float = -1
    dA_sr2: float = 1
    dB_sr1: float = -1
    dB_sr2: float = 1
    J0: float = 1
    return {
        "dA_sr1": dA_sr1,
        "dA_sr2": dA_sr2,
        "dB_sr1": dB_sr1,
        "dB_sr2": dB_sr2,
        "J0": J0,
    }



"""Transform SBML complexity into simpler representation.

Transformations:

- species and parameters
  Use simpler definition of constant parameters and dynamic variables
  SBML parameters **can** change and species **can** be constant
  Regroup these by whether they change.


- When the attribute 'initialAmount' is set, the unit of measurement associated with
    the value of 'initialAmount' is specified by the Species attribute 'substanceUnits'
- When the 'initialConcentration' attribute is set, the unit of measurement
    associated with this concentration value is {unit of amount} divided by
    {unit of size}, where the {unit of amount} is specified by the Species
    'substanceUnits' attribute, and the {unit of size} is specified by the 'units'
    attribute of the Compartment object in which the species is located
- Note that in either case, a unit of amount is involved and determined by
    the 'substanceUnits' attribute
- Note these two attributes alone do not determine the units of the species when
    the species identifier appears in a mathematical expression;
    that aspect is determined by the attribute 'hasOnlySubstanceUnits' discussed below

Additional considerations for interpreting the numerical value of a species
- Species are unique in SBML in that they have a kind of duality:
    a species identifier may stand for either
    - substance amount (meaning, a count of the number of individual entities)
    - a concentration or density (meaning, amount divided by a compartment size).
- When a species definition has a 'hasOnlySubstanceUnits' attribute value of False
    and the size of the compartment in which the species is located changes,
    the default in SBML is to assume that it is the concentration
    that must be updated to account for the size change.
- There is one exception: if the species' quantity is determined by an AssignmentRule,
    RateRule, AlgebraicRule, or an EventAssignment and the species has a
    'hasOnlySubstanceUnits' attribute value of False,
    it means that the concentration is assigned by the rule or event;
    in that case, the amount must be calculated when the compartment size changes
- (Events also require additional care in this situation, because an event with
    multiple assignments could conceivably reassign both a species quantity and a
    compartment size simultaneously.
    Please refer to the SBML specifications for the details.)

- Note that the above only matters if a species has a 'hasOnlySubstanceUnits'
    attribute value of False, meaning that the species identifier refers to a
    concentration wherever the identifier appears in a mathematical formula.
    If instead the attribute's value is True, then the identifier of the species
    always stands for an amount wherever it appears in a mathematical formula or
    is referenced by an SBML construct. In that case, there is never a question about
    whether an assignment or event is meant to affect the amount or concentration:
    it is always the amount.

- A particularly confusing situation can occur when the species has 'constant'
    attribute value of True in combination with a 'hasOnlySubstanceUnits' attribute
    value of False. Suppose this species is given a value for 'initialConcentration'.
    Does a 'constant' value of True mean that the concentration is held constant if
    the compartment size changes? No; it is still the amount that is kept constant
    across a compartment size change. The fact that the species was initialized using
    a concentration value is irrelevant.

Source: https://sbml.org/software/libsbml/5.18.0/docs/formatted/python-api/classlibsbml_1_1_species.html


"""

import logging
import math
from collections import defaultdict
from dataclasses import dataclass
from functools import reduce
from operator import mul
from typing import cast

import sympy

from pysbml.parse import data as pdata
from pysbml.transform.units import CONVERSION, PREFIXES

from . import data
from .mathml2sympy import convert_mathml

LOGGER = logging.getLogger(__name__)


def expr(x: data.Expr | sympy.Basic) -> sympy.Expr:
    return cast(sympy.Expr, x)


@dataclass
class Ctx:
    rxns_by_var: defaultdict[str, set[str]]
    ass_rules_by_var: defaultdict[str, set[str]]


def _to_sympy_types(
    x: str | float | data.Expr | sympy.Basic,
) -> data.Expr:
    if isinstance(x, str):
        return sympy.Symbol(x)
    if isinstance(x, float | int):
        return sympy.Float(x)
    return x  # type: ignore


def _div_expr(
    x: str | float | data.Expr | sympy.Basic,
    y: str | float | data.Expr | sympy.Basic,
) -> sympy.Expr:
    return _to_sympy_types(x) / _to_sympy_types(y)  # type: ignore


def _mul_expr(
    x: str | float | data.Expr | sympy.Basic,
    y: str | float | data.Expr | sympy.Basic,
) -> sympy.Expr:
    return _to_sympy_types(x) * _to_sympy_types(y)  # type: ignore


def compartment_is_valid(pmodel: pdata.Model, species: pdata.Species) -> bool:
    if (comp := species.compartment) is None:
        return False
    return bool(
        (
            pmodel.compartments[comp].size != 0
            and not math.isnan(pmodel.compartments[comp].size)
        )
        or comp in pmodel.assignment_rules
        or comp in pmodel.initial_assignments
    )


def variable_is_constant(name: str, pmodel: pdata.Model) -> bool:
    var = pmodel.variables[name]
    if var.is_constant:
        return True
    if var.has_boundary_condition:
        return name not in pmodel.rate_rules
    return False


def free_symbols(expr: sympy.Expr) -> list[str]:
    return [i.name for i in expr.free_symbols if isinstance(i, sympy.Symbol)]


def convert_units(pmodel: pdata.Model, tmodel: data.Model) -> None:
    """Replace SBML units with sympy ones."""

    for name, unit in pmodel.atomic_units.items():
        if (mapped := CONVERSION.get(name)) is None:
            LOGGER.warning("Could not map unit %s", name)
            continue
        if (prefix := PREFIXES.get(unit.exponent - 1)) is not None:
            mapped *= prefix
        tmodel.units[name] = mapped

    for name, cunit in pmodel.composite_units.items():
        tmodel.units[name] = reduce(
            mul,
            [tmodel.units.get(k, 1) for k in cunit.units],
        )


def convert_constraints(
    pmodel: pdata.Model,
    tmodel: data.Model,  # noqa: ARG001
) -> None:
    for _ in pmodel.constraints.items():
        msg = "Constraint handling not yet supported"
        raise NotImplementedError(msg)


def convert_events(
    pmodel: pdata.Model,
    tmodel: data.Model,  # noqa: ARG001
) -> None:
    for _ in pmodel.events.items():
        msg = "Event handling not yet supported"
        raise NotImplementedError(msg)


def convert_functions(pmodel: pdata.Model, tmodel: data.Model) -> None:
    for name, fn in pmodel.functions.items():
        tmodel.functions[name] = convert_mathml(fn.body, fns=tmodel.functions)


def convert_parameters(pmodel: pdata.Model, tmodel: data.Model) -> None:
    for k, par in pmodel.parameters.items():
        if par.is_constant:
            tmodel.parameters[k] = data.Parameter(
                value=sympy.Float(par.value), unit=None
            )
        else:
            tmodel.variables[k] = data.Variable(value=sympy.Float(par.value), unit=None)


def convert_compartments(pmodel: pdata.Model, tmodel: data.Model) -> None:
    for k, par in pmodel.compartments.items():
        if par.is_constant:
            tmodel.parameters[k] = data.Parameter(
                value=sympy.Float(par.size), unit=None
            )
        else:
            tmodel.variables[k] = data.Variable(value=sympy.Float(par.size), unit=None)


def convert_rules_and_initial_assignments(
    pmodel: pdata.Model, tmodel: data.Model
) -> None:
    for name, rr in pmodel.rate_rules.items():
        # Rate rules can create variables by SBML spec. Not cool
        if name not in tmodel.variables:
            tmodel.variables[name] = data.Variable(value=sympy.Float(0.0), unit=None)

        tmodel.reactions[f"d{name}"] = data.Reaction(
            expr=convert_mathml(rr.body, fns=tmodel.functions),
            stoichiometry={name: sympy.Float(1.0)},
        )

    for _ in pmodel.algebraic_rules.items():
        msg = "Algebraic rules not yet supported"
        raise NotImplementedError(msg)

    for name, ar in pmodel.assignment_rules.items():
        tmodel.derived[name] = convert_mathml(ar.body, fns=tmodel.functions)

    for name, ia in pmodel.initial_assignments.items():
        tmodel.initial_assignments[name] = convert_mathml(ia.body, fns=tmodel.functions)


def _convert_stoich_tuple(x: tuple[float, str]) -> sympy.Expr:
    factor, name = x
    return sympy.Mul(sympy.Float(factor), sympy.Symbol(name))


def convert_reactions(pmodel: pdata.Model, tmodel: data.Model) -> None:
    for name, rxn in pmodel.reactions.items():
        fn = convert_mathml(rxn.body, fns=tmodel.functions)
        stoichiometry: data.Stoichiometry = {}
        for k, v in rxn.stoichiometry.items():
            if isinstance(v, list):
                for tpl in v:
                    stoichiometry[k] = stoichiometry.get(
                        k, sympy.Float(0.0)
                    ) + _convert_stoich_tuple(tpl)  # type: ignore
            else:
                stoichiometry[k] = stoichiometry.get(k, sympy.Float(0.0)) + sympy.Float(
                    v
                )  # type: ignore

        pars_to_replace = {pn: f"{name}_{pn}" for pn in rxn.local_pars}
        fn = expr(fn.subs(pars_to_replace))
        tmodel.reactions[name] = data.Reaction(
            expr=fn,
            stoichiometry=stoichiometry,
        )

        for pn, par in rxn.local_pars.items():
            tmodel.parameters[pars_to_replace[pn]] = data.Parameter(
                value=sympy.Float(par.value), unit=None
            )


def remove_duplicate_entries(tmodel: data.Model) -> None:
    for name in tmodel.derived:
        if name in tmodel.parameters:
            del tmodel.parameters[name]
        elif name in tmodel.variables:
            del tmodel.variables[name]


def _handle_amount(
    k: str,
    compartment: str,
    init: sympy.Float,
    tmodel: data.Model,
    ctx: Ctx,
) -> None:
    """This is the default case for most tests. We are given an species in an amount.

    SBML reactions are written in a way that contains the compartment, e.g.

    | Reaction         | k1 Unit   | Unit      | Test examples |
    | ---------------- | --------- | --------- |               |
    | C * S1 * k1      | 1/s       | L * mol/s | 1             |
    | C * S1 * S2 * k1 | L/(mol*s) | L^2*mol/s | 52, 76        |
    | C * S1^2 * k1    | L/(mol*s) | L^2*mol/s | 52, 76        |


    This means the stoichiometries need to be adjusted accordingly
        1 substrate  => rxn / C
        2 substrates => rxn / C^2
    and so on

    *Counting* this out is in principle doable, but annoying.

    So let's transform the amounts into concentrations instead. This gives a somewhat
    sub-optimal representation because we need to do more work per integration step, but
    is much easier to handle.

    Steps:
      - derive concentration as amount / compartment
      - substitute amount with concentration in reactions & rules
      - remove compartment from reactions & rules
      - multiply stoichiometry by compartment to get amount again

    Initial assignment
      - initial assignments are apparently also given in units of concentration so we
        need to multiply by the compartment here
      - if the compartment also has an initial assignment, use the updated value
        (this should be handled by sorting the dependencies in our case)

    Assignment rules (676, 677, 678)
      - assigns an amount and is given an amount, nothing to fix here

    Rate rules
      - they are always given in units of concentration, so we need to multiply
        by the concentration to get an amount

    """
    tmodel.variables[k] = data.Variable(value=init, unit=None)
    tmodel.derived[k_conc := f"{k}_conc"] = _div_expr(k, compartment)

    # Fix initial assignment rule
    if (ar := tmodel.initial_assignments.get(k)) is not None:
        tmodel.initial_assignments[k] = _mul_expr(ar, compartment)

    # Fix assignment rules
    # Nothing to do here :)

    # Fix rate rule
    if (rr := tmodel.reactions.get(f"d{k}")) is not None:
        rr.stoichiometry = {k: sympy.Symbol(compartment)}

    # Fix reactions
    for rxn_name in ctx.rxns_by_var[k]:
        rxn = tmodel.reactions[rxn_name]
        rxn.expr = expr(
            rxn.expr.subs(
                {
                    compartment: 1,
                    k: k_conc,
                }
            )
        )
        if (s := rxn.stoichiometry.get(k)) is not None:
            rxn.stoichiometry[k] = _mul_expr(s, compartment)


def _handle_amount_boundary(
    k: str,
    compartment: str,
    init: sympy.Float,
    pmodel: pdata.Model,
    tmodel: data.Model,
    ctx: Ctx,
) -> None:
    """Handle amount that has a boundary condition.

    This means, per the spec (4.6.6) that the species is on the boundary of the reaction
    system, and its amount is not determined by the reactions.

    This does **not** mean, that it cannot be changed, because it **can** be changed by
    rate rules.

    """
    if k not in pmodel.rate_rules:
        tmodel.parameters[k] = data.Parameter(value=init, unit=None)
    else:
        tmodel.variables[k] = data.Variable(value=init, unit=None)
    tmodel.derived[k_conc := f"{k}_conc"] = _div_expr(k, compartment)

    # Fix initial assignment rule
    if (ar := tmodel.initial_assignments.get(k)) is not None:
        tmodel.initial_assignments[k] = _mul_expr(ar, compartment)

    # Fix assignment rules
    # Nothing to do here :)

    # Fix rate rule
    if (rr := tmodel.reactions.get(f"d{k}")) is not None:
        rr.stoichiometry = {k: sympy.Symbol(compartment)}

    # Fix reactions
    for rxn_name in ctx.rxns_by_var[k]:
        rxn = tmodel.reactions[rxn_name]

        rate = expr(rxn.expr.subs(compartment, 1))

        # FIXME: Test 1122 requires me to divide the concentration by the compartment
        # to run through. That's certainly false. What the hell is this?
        rxn.expr = expr(rate.subs(k, _div_expr(k_conc, compartment)))

        # Boundary condition means it cannot be part of the reactions system, so we
        # don't need to worry about the stoichiometry


def _handle_amount_has_substance_units(
    k: str,
    init: sympy.Float,
    tmodel: data.Model,
) -> None:
    """Handle amount with has_substance_units=True.

    According to the spec (4.6.5) the `hasOnlySubstanceUnits` allows choosing the
    meaning intended for a species' identifier when the identifier appears in
    mathematical expressions or as the subject of SBML rules or assignments.
    If the value is false the unit of measurement is a concentration or density, else
    it is always interpreted as having an amount.

    The only way that leads to legal expressions is when they don't contain a compartment
    So we can just essentially ignore everything

    """
    tmodel.variables[k] = data.Variable(value=init, unit=None)

    # Fix initial assignment
    # Nothing to do here :)

    # Fix other assignment rules
    # Nothing to do here :)

    # Fix rate rule
    # Nothing to do here :)

    # Fix reactions
    # Nothing to do here :)


def _handle_amount_boundary_has_substance_units(
    k: str,
    compartment: str,
    init: sympy.Float,
    pmodel: pdata.Model,
    tmodel: data.Model,
    ctx: Ctx,
) -> None:
    """Handle amount with boundary=True and has_substance_units=True

    A boundary condition means, per the spec (4.6.6), that the species is on the
    boundary of the reaction system, and its amount is not determined by the reactions.

    According to the spec (4.6.5) the `hasOnlySubstanceUnits` allows choosing the
    meaning intended for a species' identifier when the identifier appears in
    mathematical expressions or as the subject of SBML rules or assignments.
    If the value is true the unit of measurement is always interpreted as an amount.

    In test 1123 we have
        S1: only substance units = False
        S3: only substance units = True

        rxn J0 = S3 / 10 with stoichiometry S1: -1.0

    Since by our logic S1 will give itself the stoichiometry S1: -comp, we need to ignore
    the hasOnlySubstanceUnits logic and insert the concentration into the reaction

    FIXME: why does this happen with the boundary, but not without?
    """
    if k not in pmodel.rate_rules:
        tmodel.parameters[k] = data.Parameter(value=init, unit=None)
    else:
        tmodel.variables[k] = data.Variable(value=init, unit=None)

    # We need the concentration of the boundary species in reactions
    k_conc = f"{k}_conc"
    tmodel.derived[k_conc] = _div_expr(k, compartment)

    # Fix reactions
    for rxn_name in ctx.rxns_by_var[k]:
        rxn = tmodel.reactions[rxn_name]
        rxn.expr = expr(rxn.expr.subs(k, k_conc))


def _handle_conc(
    k: str,
    compartment: str,
    init: sympy.Float,
    pmodel: pdata.Model,
    tmodel: data.Model,
    ctx: Ctx,
) -> None:
    """Handle species given as a concentration.

    By default this is straightforward, as species are interpreted as concentrations in
    all expressions (except if hasOnlySubstanceUnits=True).

    There is a slight complication though if a compartment is changing. In that case it
    is easier to also create the amount.
    """
    if pmodel.compartments[compartment].is_constant:
        tmodel.variables[k] = data.Variable(value=init, unit=None)

        # Fix initial assignment rule
        if (comp := tmodel.initial_assignments.get(compartment)) is not None:
            tmodel.initial_assignments[k] = _mul_expr(init, comp)

        # Fix assignment rule
        # Nothing to do here :)

        # Fix rate rule
        # Nothing to do here :)

        # Fix reactions
        for rxn_name in ctx.rxns_by_var[k]:
            rxn = tmodel.reactions[rxn_name]
            rxn.expr = expr(rxn.expr.subs(compartment, 1))
    else:
        tmodel.variables[k_amount := f"{k}_amount"] = data.Variable(
            value=init, unit=None
        )
        tmodel.derived[k] = _div_expr(k_amount, compartment)

        # Fix initial assignment rule
        if (ar := tmodel.initial_assignments.get(k)) is not None:
            tmodel.initial_assignments[k_amount] = _mul_expr(ar, compartment)
        else:
            tmodel.initial_assignments[k_amount] = _mul_expr(init, compartment)

        # Fix assignment rules
        # Nothing to do here :)

        # Fix rate rule
        # Nothing to do here :)

        # Fix reactions
        for rxn_name in ctx.rxns_by_var[k]:
            rxn = tmodel.reactions[rxn_name]
            if k in rxn.stoichiometry:
                rxn.stoichiometry[k_amount] = rxn.stoichiometry.pop(k)


def _handle_conc_boundary(
    k: str,
    compartment: str,
    init: sympy.Float,
    tmodel: data.Model,
    ctx: Ctx,
) -> None:
    """ """
    tmodel.variables[k_conc := f"{k}_conc"] = data.Variable(value=init, unit=None)
    tmodel.derived[k] = _mul_expr(k_conc, compartment)

    # Fix initial assignment rule
    if (ia := tmodel.initial_assignments.pop(k, None)) is not None:
        tmodel.initial_assignments[k_conc] = ia

    # Fix assignment rules
    for dname in ctx.ass_rules_by_var[k]:
        if dname == k:
            tmodel.derived[k_conc] = expr(tmodel.derived.pop(dname).subs(k, k_conc))
        else:
            tmodel.derived[dname] = expr(tmodel.derived[dname].subs(k, k_conc))

    # Fix rate rule
    if (rr := tmodel.reactions.pop(f"d{k}", None)) is not None:
        tmodel.reactions[f"d{k_conc}"] = rr

        if rr.stoichiometry.get(k) is not None:
            rr.stoichiometry[k_conc] = rr.stoichiometry.pop(k)

    # Fix reactions
    # Nothing to do here, boundary species cannot have reactions :)


def _handle_conc_has_substance_units(
    k: str,
    compartment: str,
    init: sympy.Float,
    tmodel: data.Model,
) -> None:
    """Handle a species given in a concentration that is always intepreted as an amount.

    Sigh. Ok, so let's just multiply the concentration by the compartment and call it a
    day.

    """
    tmodel.variables[k] = data.Variable(value=init, unit=None)
    if (ia := tmodel.initial_assignments.get(k)) is not None:
        tmodel.initial_assignments[k] = _mul_expr(ia, compartment)
    else:
        tmodel.initial_assignments[k] = _mul_expr(init, compartment)

    # Fix initial assignment rule
    # Nothing to do here :)

    # Fix assignment rules
    # Nothing to do here :)

    # Fix rate rule
    # Nothing to do here :)

    # Fix reactions
    # Nothing to do here :)


def _handle_conc_boundary_has_substance_units(
    k: str,
    compartment: str,
    init: sympy.Float,
    tmodel: data.Model,
    ctx: Ctx,
) -> None:
    """Handle a species given in a concentration that is always intepreted as an amount
    as well as a boundary.

    Sigh. Ok, so let's just multiply the concentration by the compartment and call it a
    day.

    """
    tmodel.variables[k] = data.Variable(value=init, unit=None)
    if (ia := tmodel.initial_assignments.get(k)) is not None:
        tmodel.initial_assignments[k] = _mul_expr(ia, compartment)
    else:
        tmodel.initial_assignments[k] = _mul_expr(init, compartment)

    # Fix initial assignment rule
    # Nothing to do here :)

    # Fix assignment rules
    for dname in ctx.ass_rules_by_var[k]:
        tmodel.derived[dname] = _mul_expr(tmodel.derived[dname], compartment)

    # Fix rate rule
    # Nothing to do here :)

    # Fix reactions
    # Nothing to do here :)


def _transform_species(
    k: str,
    species: pdata.Species,
    pmodel: pdata.Model,
    tmodel: data.Model,
    ctx: Ctx,
) -> None:
    """Separate species into parameters and variables and substitute correct version
    in reactions, rules and initial assignments if necessary.
    """
    if species.conversion_factor is not None:
        raise NotImplementedError

    init = sympy.Float(
        init
        if (init := species.initial_amount) is not None
        or (init := species.initial_concentration) is not None
        else 0.0
    )

    # Now start making case distinctions
    # Easiest is to check first if the compartment is valid
    # If not, our life is significantly easier, because there really are just two choices
    if not compartment_is_valid(pmodel, species=species):
        if variable_is_constant(k, pmodel):
            tmodel.parameters[k] = data.Parameter(value=init, unit=None)
            return

        tmodel.variables[k] = data.Variable(value=init, unit=None)
        return

    # Compartment is valid as in exists and is non-zero / nan
    compartment = cast(str, species.compartment)

    # Now the garbage begins
    # I'm going to do something disgusting by now and write out every case explicitly
    # in a nested way to see the entire decision tree
    # I know this is bad code, I'll refactor it later

    # Let's separate next by is_concentration / is_amount / is_to_be_determined :')
    # Because of course some of them are annotated without either conc or amount

    # We have an amount here
    if species.initial_amount is not None:
        if species.has_only_substance_units:
            if species.has_boundary_condition:
                LOGGER.debug("Species %s: amount | True | True", k)
                _handle_amount_boundary_has_substance_units(
                    k=k,
                    compartment=compartment,
                    init=init,
                    pmodel=pmodel,
                    tmodel=tmodel,
                    ctx=ctx,
                )

            else:
                LOGGER.debug("Species %s amount | True | False", k)
                _handle_amount_has_substance_units(
                    k=k,
                    init=init,
                    tmodel=tmodel,
                )

        else:  # noqa: PLR5501
            if species.has_boundary_condition:
                LOGGER.debug("Species %s: amount | False | True", k)
                _handle_amount_boundary(
                    k=k,
                    compartment=compartment,
                    init=init,
                    pmodel=pmodel,
                    tmodel=tmodel,
                    ctx=ctx,
                )

            else:
                LOGGER.debug("Species %s: amount | False | False", k)
                _handle_amount(
                    k=k,
                    compartment=compartment,
                    init=init,
                    tmodel=tmodel,
                    ctx=ctx,
                )

    # We have a concentration here
    elif species.initial_concentration is not None:
        if variable_is_constant(k, pmodel):
            tmodel.parameters[k] = data.Parameter(value=init, unit=None)
            return

        # If it IS a concentration but has only substance units
        # is set, we have to multiply it by the compartment initially
        if species.has_only_substance_units:
            if species.has_boundary_condition:
                LOGGER.debug("Species %s: | conc | True | True", k)
                _handle_conc_boundary_has_substance_units(
                    k=k,
                    compartment=compartment,
                    init=init,
                    tmodel=tmodel,
                    ctx=ctx,
                )

            else:
                LOGGER.debug("Species %s: | conc | True | False", k)
                _handle_conc_has_substance_units(
                    k=k,
                    compartment=compartment,
                    init=init,
                    tmodel=tmodel,
                )

        else:  # noqa: PLR5501
            if species.has_boundary_condition:
                LOGGER.debug("Species %s: | conc | False | True", k)
                _handle_conc_boundary(
                    k=k,
                    compartment=compartment,
                    init=init,
                    tmodel=tmodel,
                    ctx=ctx,
                )

            else:
                LOGGER.debug("Species %s: | conc | False | False", k)
                _handle_conc(
                    k=k,
                    compartment=compartment,
                    init=init,
                    pmodel=pmodel,
                    tmodel=tmodel,
                    ctx=ctx,
                )

    # Now BOTH of them are None, the whackest case of them all. If you think you can
    # figure out if it is a concentration or amount just by looking at species
    # and compartments, boy do I have a surprise for you :)
    # The documentation states you can use species.has_only_substance_units for this
    # which is false, as test cases 676 and 688 demonstrate. There S1 sets this as false
    # still it is being used as a concentration or an amount respectively
    else:
        is_concentration = False
        for rxn_name in ctx.rxns_by_var[k]:
            reaction = pmodel.reactions[rxn_name]
            targets = {i.name for i in reaction.args} | set(reaction.stoichiometry)
            for other in targets:
                if (
                    var := pmodel.variables.get(other)
                ) is not None and var.initial_concentration is not None:
                    is_concentration = True
                    break
        # Inject concentration and run the whole thing again to avoid
        # duplicating all those conditions
        if is_concentration:
            pmodel.variables[k].initial_concentration = 0.0
            _transform_species(k, species, pmodel, tmodel, ctx)

        # Fall back to interpretation as amount if no evidence for concentration
        # was found
        else:
            # test 676 assumes amount
            # test 1513 assumes concentration??
            pmodel.variables[k].initial_amount = 0.0
            _transform_species(k, species, pmodel, tmodel, ctx)


def transform_species(pmodel: pdata.Model, tmodel: data.Model, ctx: Ctx) -> None:
    LOGGER.debug("Species name | type | only subs. | boundary cond.")
    for k, var in pmodel.variables.items():
        _transform_species(k, var, pmodel, tmodel, ctx=ctx)


def transform(doc: pdata.Document) -> data.Model:
    for plugin in doc.plugins:
        if plugin.name == "comp":
            msg = "Comp package not yet supported."
            raise NotImplementedError(msg)

    pmodel = doc.model
    if pmodel.conversion_factor is not None:
        msg = "Conversion factors not yet supported"
        raise NotImplementedError(msg)

    ctx = Ctx(rxns_by_var=defaultdict(set), ass_rules_by_var=defaultdict(set))
    for name, rxn in pmodel.reactions.items():
        for arg in rxn.args:
            ctx.rxns_by_var[arg.name].add(name)
        for arg in rxn.stoichiometry:
            ctx.rxns_by_var[arg].add(name)
    for name, rule in pmodel.assignment_rules.items():
        ctx.ass_rules_by_var[name].add(name)
        for arg in rule.args:
            ctx.ass_rules_by_var[arg.name].add(name)

    tmodel = data.Model(name=pmodel.name)  # type: ignore
    convert_units(pmodel=pmodel, tmodel=tmodel)
    convert_parameters(pmodel=pmodel, tmodel=tmodel)
    convert_compartments(pmodel=pmodel, tmodel=tmodel)
    convert_constraints(pmodel=pmodel, tmodel=tmodel)
    convert_events(pmodel=pmodel, tmodel=tmodel)
    convert_functions(pmodel=pmodel, tmodel=tmodel)
    convert_rules_and_initial_assignments(pmodel=pmodel, tmodel=tmodel)
    convert_reactions(pmodel=pmodel, tmodel=tmodel)

    # Do the heavy lifting here
    transform_species(pmodel=pmodel, tmodel=tmodel, ctx=ctx)
    remove_duplicate_entries(tmodel=tmodel)
    return tmodel



time: float = 0.0
C: float = 2.50000000000000
S2: float = 0.150000000000000

# Initial assignments
k1 = 0.750000000000000
S1 = 0.133333333333333 * k1
reaction1 = S1 * k1
y0 = [S1, S2]
variable_names = ["S1", "S2"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, S2 = variables
    k1: float = 0.750000000000000
    reaction1: float = S1 * k1
    dS1dt: float = -reaction1
    dS2dt: float = reaction1
    return dS1dt, dS2dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, S2 = variables
    k1: float = 0.750000000000000
    reaction1: float = S1 * k1
    return {
        "k1": k1,
        "reaction1": reaction1,
    }



import math

time: float = 0.0
C: float = 1.00000000000000
S0_stoich: float = 0.0
S1_stoich: float = 0.0
S2_stoich: float = 0.0
S3_stoich: float = 0.0
S4_stoich: float = 0.0
S5_stoich: float = 0.0
S6_stoich: float = 0.0
S7_stoich: float = 0.0
S8_stoich: float = 0.0
S9_stoich: float = 0.0
S10_stoich: float = 0.0
S11_stoich: float = 0.0
S12_stoich: float = 0.0
S13_stoich: float = 0.0
S14_stoich: float = 0.0
S15_stoich: float = 0.0
S16_stoich: float = 0.0
S17_stoich: float = 0.0
S18_stoich: float = 0.0
S19_stoich: float = 0.0
S20_stoich: float = 0.0
S21_stoich: float = 0.0
S22_stoich: float = 0.0
S23_stoich: float = 0.0
S24_stoich: float = 0.0
S25_stoich: float = 0.0
S26_stoich: float = 0.0
S27_stoich: float = 0.0
S28_stoich: float = 0.0
S29_stoich: float = 0.0
S30_stoich: float = 0.0
S31_stoich: float = 0.0
S32_stoich: float = 0.0
S33_stoich: float = 0.0
S34_stoich: float = 0.0
S35_stoich: float = 0.0
S36_stoich: float = 0.0
S37_stoich: float = 0.0
S38_stoich: float = 0.0
S39_stoich: float = 0.0
S40_stoich: float = 0.0
S41_stoich: float = 0.0
S42_stoich: float = 0.0
S43_stoich: float = 0.0
S44_stoich: float = 0.0
S45_stoich: float = 0.0
S46_stoich: float = 0.0
S47_stoich: float = 0.0
S48_stoich: float = 0.0
S49_stoich: float = 0.0
S50_stoich: float = 0.0
S51_stoich: float = 0.0
S0: float = 0.0
S1: float = 0.0
S2: float = 0.0
S3: float = 0.0
S4: float = 0.0
S5: float = 0.0
S6: float = 0.0
S7: float = 0.0
S8: float = 0.0
S9: float = 0.0
S10: float = 0.0
S11: float = 0.0
S12: float = 0.0
S13: float = 0.0
S14: float = 0.0
S15: float = 0.0
S16: float = 0.0
S17: float = 0.0
S18: float = 0.0
S19: float = 0.0
S20: float = 0.0
S21: float = 0.0
S22: float = 0.0
S23: float = 0.0
S24: float = 0.0
S25: float = 0.0
S26: float = 0.0
S27: float = 0.0
S28: float = 0.0
S29: float = 0.0
S30: float = 0.0
S31: float = 0.0
S32: float = 0.0
S33: float = 0.0
S34: float = 0.0
S35: float = 0.0
S36: float = 0.0
S37: float = 0.0
S38: float = 0.0
S39: float = 0.0
S40: float = 0.0
S41: float = 0.0
S42: float = 0.0
S43: float = 0.0
S44: float = 0.0
S45: float = 0.0
S46: float = 0.0
S47: float = 0.0
S48: float = 0.0
S49: float = 0.0
S50: float = 0.0
S51: float = 0.0

# Initial assignments
dS0_stoich = math.e
dS1_stoich = math.exp(math.e)
dS2_stoich = 1
dS3_stoich = 1
dS4_stoich = math.pi
dS5_stoich = 1.04719755119660
dS6_stoich = (1 / 2) * math.pi
dS7_stoich = -0.523598775598299
dS8_stoich = 1.22777238637419
dS9_stoich = -1.43067687253053
dS10_stoich = 1
dS11_stoich = 4
dS12_stoich = -4
dS13_stoich = -0.947721602131112
dS14_stoich = 0.975897449330606
dS15_stoich = 1
dS16_stoich = math.e
dS17_stoich = 2.15976625378492
dS18_stoich = -5
dS19_stoich = 9
dS20_stoich = -1.60943791243410
dS21_stoich = 0
dS22_stoich = -1.6094379124341 / math.log(10)
dS23_stoich = 0
dS24_stoich = 1
dS25_stoich = 4
dS26_stoich = 5.56228172775440
dS27_stoich = 16
dS28_stoich = 9.61000000000000
dS29_stoich = 2
dS30_stoich = 2.72029410174709
dS31_stoich = 0.863209366648874
dS32_stoich = 0
dS33_stoich = 0.373876664830236
dS34_stoich = 0
dS35_stoich = 2.01433821447683
dS36_stoich = -math.tan(6)
dS37_stoich = 1.13949392732455
dS38_stoich = -1.02298638367130
dS39_stoich = 4.93315487558689
dS40_stoich = 0.304520293447143
dS41_stoich = 2.82831545788997
dS42_stoich = 1.12099946637054
dS43_stoich = 1.14109666064347
dS44_stoich = -1.47112767430373
dS45_stoich = math.asinh(99)
dS46_stoich = 0.802881971289134
dS47_stoich = -0.867300527694053
dS48_stoich = 1.51330668842682
dS49_stoich = 5.29834236561059
dS50_stoich = 1
dS51_stoich = 0
J0 = 1
J1 = 1
J2 = 1
J3 = 1
J4 = 1
J5 = 1
J6 = 1
J7 = 1
J8 = 1
J9 = 1
J10 = 1
J11 = 1
J12 = 1
J13 = 1
J14 = 1
J15 = 1
J16 = 1
J17 = 1
J18 = 1
J19 = 1
J20 = 1
J21 = 1
J22 = 1
J23 = 1
J24 = 1
J25 = 1
J26 = 1
J27 = 1
J28 = 1
J29 = 1
J30 = 1
J31 = 1
J32 = 1
J33 = 1
J34 = 1
J35 = 1
J36 = 1
J37 = 1
J38 = 1
J39 = 1
J40 = 1
J41 = 1
J42 = 1
J43 = 1
J44 = 1
J45 = 1
J46 = 1
J47 = 1
J48 = 1
J49 = 1
J50 = 1
J51 = 1
y0 = [
    S0,
    S0_stoich,
    S1,
    S10,
    S10_stoich,
    S11,
    S11_stoich,
    S12,
    S12_stoich,
    S13,
    S13_stoich,
    S14,
    S14_stoich,
    S15,
    S15_stoich,
    S16,
    S16_stoich,
    S17,
    S17_stoich,
    S18,
    S18_stoich,
    S19,
    S19_stoich,
    S1_stoich,
    S2,
    S20,
    S20_stoich,
    S21,
    S21_stoich,
    S22,
    S22_stoich,
    S23,
    S23_stoich,
    S24,
    S24_stoich,
    S25,
    S25_stoich,
    S26,
    S26_stoich,
    S27,
    S27_stoich,
    S28,
    S28_stoich,
    S29,
    S29_stoich,
    S2_stoich,
    S3,
    S30,
    S30_stoich,
    S31,
    S31_stoich,
    S32,
    S32_stoich,
    S33,
    S33_stoich,
    S34,
    S34_stoich,
    S35,
    S35_stoich,
    S36,
    S36_stoich,
    S37,
    S37_stoich,
    S38,
    S38_stoich,
    S39,
    S39_stoich,
    S3_stoich,
    S4,
    S40,
    S40_stoich,
    S41,
    S41_stoich,
    S42,
    S42_stoich,
    S43,
    S43_stoich,
    S44,
    S44_stoich,
    S45,
    S45_stoich,
    S46,
    S46_stoich,
    S47,
    S47_stoich,
    S48,
    S48_stoich,
    S49,
    S49_stoich,
    S4_stoich,
    S5,
    S50,
    S50_stoich,
    S51,
    S51_stoich,
    S5_stoich,
    S6,
    S6_stoich,
    S7,
    S7_stoich,
    S8,
    S8_stoich,
    S9,
    S9_stoich,
]
variable_names = [
    "S0",
    "S0_stoich",
    "S1",
    "S10",
    "S10_stoich",
    "S11",
    "S11_stoich",
    "S12",
    "S12_stoich",
    "S13",
    "S13_stoich",
    "S14",
    "S14_stoich",
    "S15",
    "S15_stoich",
    "S16",
    "S16_stoich",
    "S17",
    "S17_stoich",
    "S18",
    "S18_stoich",
    "S19",
    "S19_stoich",
    "S1_stoich",
    "S2",
    "S20",
    "S20_stoich",
    "S21",
    "S21_stoich",
    "S22",
    "S22_stoich",
    "S23",
    "S23_stoich",
    "S24",
    "S24_stoich",
    "S25",
    "S25_stoich",
    "S26",
    "S26_stoich",
    "S27",
    "S27_stoich",
    "S28",
    "S28_stoich",
    "S29",
    "S29_stoich",
    "S2_stoich",
    "S3",
    "S30",
    "S30_stoich",
    "S31",
    "S31_stoich",
    "S32",
    "S32_stoich",
    "S33",
    "S33_stoich",
    "S34",
    "S34_stoich",
    "S35",
    "S35_stoich",
    "S36",
    "S36_stoich",
    "S37",
    "S37_stoich",
    "S38",
    "S38_stoich",
    "S39",
    "S39_stoich",
    "S3_stoich",
    "S4",
    "S40",
    "S40_stoich",
    "S41",
    "S41_stoich",
    "S42",
    "S42_stoich",
    "S43",
    "S43_stoich",
    "S44",
    "S44_stoich",
    "S45",
    "S45_stoich",
    "S46",
    "S46_stoich",
    "S47",
    "S47_stoich",
    "S48",
    "S48_stoich",
    "S49",
    "S49_stoich",
    "S4_stoich",
    "S5",
    "S50",
    "S50_stoich",
    "S51",
    "S51_stoich",
    "S5_stoich",
    "S6",
    "S6_stoich",
    "S7",
    "S7_stoich",
    "S8",
    "S8_stoich",
    "S9",
    "S9_stoich",
]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    (
        S0,
        S0_stoich,
        S1,
        S10,
        S10_stoich,
        S11,
        S11_stoich,
        S12,
        S12_stoich,
        S13,
        S13_stoich,
        S14,
        S14_stoich,
        S15,
        S15_stoich,
        S16,
        S16_stoich,
        S17,
        S17_stoich,
        S18,
        S18_stoich,
        S19,
        S19_stoich,
        S1_stoich,
        S2,
        S20,
        S20_stoich,
        S21,
        S21_stoich,
        S22,
        S22_stoich,
        S23,
        S23_stoich,
        S24,
        S24_stoich,
        S25,
        S25_stoich,
        S26,
        S26_stoich,
        S27,
        S27_stoich,
        S28,
        S28_stoich,
        S29,
        S29_stoich,
        S2_stoich,
        S3,
        S30,
        S30_stoich,
        S31,
        S31_stoich,
        S32,
        S32_stoich,
        S33,
        S33_stoich,
        S34,
        S34_stoich,
        S35,
        S35_stoich,
        S36,
        S36_stoich,
        S37,
        S37_stoich,
        S38,
        S38_stoich,
        S39,
        S39_stoich,
        S3_stoich,
        S4,
        S40,
        S40_stoich,
        S41,
        S41_stoich,
        S42,
        S42_stoich,
        S43,
        S43_stoich,
        S44,
        S44_stoich,
        S45,
        S45_stoich,
        S46,
        S46_stoich,
        S47,
        S47_stoich,
        S48,
        S48_stoich,
        S49,
        S49_stoich,
        S4_stoich,
        S5,
        S50,
        S50_stoich,
        S51,
        S51_stoich,
        S5_stoich,
        S6,
        S6_stoich,
        S7,
        S7_stoich,
        S8,
        S8_stoich,
        S9,
        S9_stoich,
    ) = variables
    dS0_stoich: float = math.e
    dS1_stoich: float = math.exp(math.e)
    dS2_stoich: float = 1
    dS3_stoich: float = 1
    dS4_stoich: float = math.pi
    dS5_stoich: float = 1.04719755119660
    dS6_stoich: float = (1 / 2) * math.pi
    dS7_stoich: float = -0.523598775598299
    dS8_stoich: float = 1.22777238637419
    dS9_stoich: float = -1.43067687253053
    dS10_stoich: float = 1
    dS11_stoich: float = 4
    dS12_stoich: float = -4
    dS13_stoich: float = -0.947721602131112
    dS14_stoich: float = 0.975897449330606
    dS15_stoich: float = 1
    dS16_stoich: float = math.e
    dS17_stoich: float = 2.15976625378492
    dS18_stoich: float = -5
    dS19_stoich: float = 9
    dS20_stoich: float = -1.60943791243410
    dS21_stoich: float = 0
    dS22_stoich: float = -1.6094379124341 / math.log(10)
    dS23_stoich: float = 0
    dS24_stoich: float = 1
    dS25_stoich: float = 4
    dS26_stoich: float = 5.56228172775440
    dS27_stoich: float = 16
    dS28_stoich: float = 9.61000000000000
    dS29_stoich: float = 2
    dS30_stoich: float = 2.72029410174709
    dS31_stoich: float = 0.863209366648874
    dS32_stoich: float = 0
    dS33_stoich: float = 0.373876664830236
    dS34_stoich: float = 0
    dS35_stoich: float = 2.01433821447683
    dS36_stoich: float = -math.tan(6)
    dS37_stoich: float = 1.13949392732455
    dS38_stoich: float = -1.02298638367130
    dS39_stoich: float = 4.93315487558689
    dS40_stoich: float = 0.304520293447143
    dS41_stoich: float = 2.82831545788997
    dS42_stoich: float = 1.12099946637054
    dS43_stoich: float = 1.14109666064347
    dS44_stoich: float = -1.47112767430373
    dS45_stoich: float = math.asinh(99)
    dS46_stoich: float = 0.802881971289134
    dS47_stoich: float = -0.867300527694053
    dS48_stoich: float = 1.51330668842682
    dS49_stoich: float = 5.29834236561059
    dS50_stoich: float = 1
    dS51_stoich: float = 0
    J0: float = 1
    J1: float = 1
    J2: float = 1
    J3: float = 1
    J4: float = 1
    J5: float = 1
    J6: float = 1
    J7: float = 1
    J8: float = 1
    J9: float = 1
    J10: float = 1
    J11: float = 1
    J12: float = 1
    J13: float = 1
    J14: float = 1
    J15: float = 1
    J16: float = 1
    J17: float = 1
    J18: float = 1
    J19: float = 1
    J20: float = 1
    J21: float = 1
    J22: float = 1
    J23: float = 1
    J24: float = 1
    J25: float = 1
    J26: float = 1
    J27: float = 1
    J28: float = 1
    J29: float = 1
    J30: float = 1
    J31: float = 1
    J32: float = 1
    J33: float = 1
    J34: float = 1
    J35: float = 1
    J36: float = 1
    J37: float = 1
    J38: float = 1
    J39: float = 1
    J40: float = 1
    J41: float = 1
    J42: float = 1
    J43: float = 1
    J44: float = 1
    J45: float = 1
    J46: float = 1
    J47: float = 1
    J48: float = 1
    J49: float = 1
    J50: float = 1
    J51: float = 1
    dS0_stoichdt: float = dS0_stoich
    dS1_stoichdt: float = dS1_stoich
    dS2_stoichdt: float = dS2_stoich
    dS3_stoichdt: float = dS3_stoich
    dS4_stoichdt: float = dS4_stoich
    dS5_stoichdt: float = dS5_stoich
    dS6_stoichdt: float = dS6_stoich
    dS7_stoichdt: float = dS7_stoich
    dS8_stoichdt: float = dS8_stoich
    dS9_stoichdt: float = dS9_stoich
    dS10_stoichdt: float = dS10_stoich
    dS11_stoichdt: float = dS11_stoich
    dS12_stoichdt: float = dS12_stoich
    dS13_stoichdt: float = dS13_stoich
    dS14_stoichdt: float = dS14_stoich
    dS15_stoichdt: float = dS15_stoich
    dS16_stoichdt: float = dS16_stoich
    dS17_stoichdt: float = dS17_stoich
    dS18_stoichdt: float = dS18_stoich
    dS19_stoichdt: float = dS19_stoich
    dS20_stoichdt: float = dS20_stoich
    dS21_stoichdt: float = dS21_stoich
    dS22_stoichdt: float = dS22_stoich
    dS23_stoichdt: float = dS23_stoich
    dS24_stoichdt: float = dS24_stoich
    dS25_stoichdt: float = dS25_stoich
    dS26_stoichdt: float = dS26_stoich
    dS27_stoichdt: float = dS27_stoich
    dS28_stoichdt: float = dS28_stoich
    dS29_stoichdt: float = dS29_stoich
    dS30_stoichdt: float = dS30_stoich
    dS31_stoichdt: float = dS31_stoich
    dS32_stoichdt: float = dS32_stoich
    dS33_stoichdt: float = dS33_stoich
    dS34_stoichdt: float = dS34_stoich
    dS35_stoichdt: float = dS35_stoich
    dS36_stoichdt: float = dS36_stoich
    dS37_stoichdt: float = dS37_stoich
    dS38_stoichdt: float = dS38_stoich
    dS39_stoichdt: float = dS39_stoich
    dS40_stoichdt: float = dS40_stoich
    dS41_stoichdt: float = dS41_stoich
    dS42_stoichdt: float = dS42_stoich
    dS43_stoichdt: float = dS43_stoich
    dS44_stoichdt: float = dS44_stoich
    dS45_stoichdt: float = dS45_stoich
    dS46_stoichdt: float = dS46_stoich
    dS47_stoichdt: float = dS47_stoich
    dS48_stoichdt: float = dS48_stoich
    dS49_stoichdt: float = dS49_stoich
    dS50_stoichdt: float = dS50_stoich
    dS51_stoichdt: float = dS51_stoich
    dS0dt: float = J0 * S0_stoich
    dS1dt: float = J1 * S1_stoich
    dS2dt: float = J2 * S2_stoich
    dS3dt: float = J3 * S3_stoich
    dS4dt: float = J4 * S4_stoich
    dS5dt: float = J5 * S5_stoich
    dS6dt: float = J6 * S6_stoich
    dS7dt: float = J7 * S7_stoich
    dS8dt: float = J8 * S8_stoich
    dS9dt: float = J9 * S9_stoich
    dS10dt: float = J10 * S10_stoich
    dS11dt: float = J11 * S11_stoich
    dS12dt: float = J12 * S12_stoich
    dS13dt: float = J13 * S13_stoich
    dS14dt: float = J14 * S14_stoich
    dS15dt: float = J15 * S15_stoich
    dS16dt: float = J16 * S16_stoich
    dS17dt: float = J17 * S17_stoich
    dS18dt: float = J18 * S18_stoich
    dS19dt: float = J19 * S19_stoich
    dS20dt: float = J20 * S20_stoich
    dS21dt: float = J21 * S21_stoich
    dS22dt: float = J22 * S22_stoich
    dS23dt: float = J23 * S23_stoich
    dS24dt: float = J24 * S24_stoich
    dS25dt: float = J25 * S25_stoich
    dS26dt: float = J26 * S26_stoich
    dS27dt: float = J27 * S27_stoich
    dS28dt: float = J28 * S28_stoich
    dS29dt: float = J29 * S29_stoich
    dS30dt: float = J30 * S30_stoich
    dS31dt: float = J31 * S31_stoich
    dS32dt: float = J32 * S32_stoich
    dS33dt: float = J33 * S33_stoich
    dS34dt: float = J34 * S34_stoich
    dS35dt: float = J35 * S35_stoich
    dS36dt: float = J36 * S36_stoich
    dS37dt: float = J37 * S37_stoich
    dS38dt: float = J38 * S38_stoich
    dS39dt: float = J39 * S39_stoich
    dS40dt: float = J40 * S40_stoich
    dS41dt: float = J41 * S41_stoich
    dS42dt: float = J42 * S42_stoich
    dS43dt: float = J43 * S43_stoich
    dS44dt: float = J44 * S44_stoich
    dS45dt: float = J45 * S45_stoich
    dS46dt: float = J46 * S46_stoich
    dS47dt: float = J47 * S47_stoich
    dS48dt: float = J48 * S48_stoich
    dS49dt: float = J49 * S49_stoich
    dS50dt: float = J50 * S50_stoich
    dS51dt: float = J51 * S51_stoich
    return (
        dS0dt,
        dS0_stoichdt,
        dS1dt,
        dS10dt,
        dS10_stoichdt,
        dS11dt,
        dS11_stoichdt,
        dS12dt,
        dS12_stoichdt,
        dS13dt,
        dS13_stoichdt,
        dS14dt,
        dS14_stoichdt,
        dS15dt,
        dS15_stoichdt,
        dS16dt,
        dS16_stoichdt,
        dS17dt,
        dS17_stoichdt,
        dS18dt,
        dS18_stoichdt,
        dS19dt,
        dS19_stoichdt,
        dS1_stoichdt,
        dS2dt,
        dS20dt,
        dS20_stoichdt,
        dS21dt,
        dS21_stoichdt,
        dS22dt,
        dS22_stoichdt,
        dS23dt,
        dS23_stoichdt,
        dS24dt,
        dS24_stoichdt,
        dS25dt,
        dS25_stoichdt,
        dS26dt,
        dS26_stoichdt,
        dS27dt,
        dS27_stoichdt,
        dS28dt,
        dS28_stoichdt,
        dS29dt,
        dS29_stoichdt,
        dS2_stoichdt,
        dS3dt,
        dS30dt,
        dS30_stoichdt,
        dS31dt,
        dS31_stoichdt,
        dS32dt,
        dS32_stoichdt,
        dS33dt,
        dS33_stoichdt,
        dS34dt,
        dS34_stoichdt,
        dS35dt,
        dS35_stoichdt,
        dS36dt,
        dS36_stoichdt,
        dS37dt,
        dS37_stoichdt,
        dS38dt,
        dS38_stoichdt,
        dS39dt,
        dS39_stoichdt,
        dS3_stoichdt,
        dS4dt,
        dS40dt,
        dS40_stoichdt,
        dS41dt,
        dS41_stoichdt,
        dS42dt,
        dS42_stoichdt,
        dS43dt,
        dS43_stoichdt,
        dS44dt,
        dS44_stoichdt,
        dS45dt,
        dS45_stoichdt,
        dS46dt,
        dS46_stoichdt,
        dS47dt,
        dS47_stoichdt,
        dS48dt,
        dS48_stoichdt,
        dS49dt,
        dS49_stoichdt,
        dS4_stoichdt,
        dS5dt,
        dS50dt,
        dS50_stoichdt,
        dS51dt,
        dS51_stoichdt,
        dS5_stoichdt,
        dS6dt,
        dS6_stoichdt,
        dS7dt,
        dS7_stoichdt,
        dS8dt,
        dS8_stoichdt,
        dS9dt,
        dS9_stoichdt,
    )


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    (
        S0,
        S0_stoich,
        S1,
        S10,
        S10_stoich,
        S11,
        S11_stoich,
        S12,
        S12_stoich,
        S13,
        S13_stoich,
        S14,
        S14_stoich,
        S15,
        S15_stoich,
        S16,
        S16_stoich,
        S17,
        S17_stoich,
        S18,
        S18_stoich,
        S19,
        S19_stoich,
        S1_stoich,
        S2,
        S20,
        S20_stoich,
        S21,
        S21_stoich,
        S22,
        S22_stoich,
        S23,
        S23_stoich,
        S24,
        S24_stoich,
        S25,
        S25_stoich,
        S26,
        S26_stoich,
        S27,
        S27_stoich,
        S28,
        S28_stoich,
        S29,
        S29_stoich,
        S2_stoich,
        S3,
        S30,
        S30_stoich,
        S31,
        S31_stoich,
        S32,
        S32_stoich,
        S33,
        S33_stoich,
        S34,
        S34_stoich,
        S35,
        S35_stoich,
        S36,
        S36_stoich,
        S37,
        S37_stoich,
        S38,
        S38_stoich,
        S39,
        S39_stoich,
        S3_stoich,
        S4,
        S40,
        S40_stoich,
        S41,
        S41_stoich,
        S42,
        S42_stoich,
        S43,
        S43_stoich,
        S44,
        S44_stoich,
        S45,
        S45_stoich,
        S46,
        S46_stoich,
        S47,
        S47_stoich,
        S48,
        S48_stoich,
        S49,
        S49_stoich,
        S4_stoich,
        S5,
        S50,
        S50_stoich,
        S51,
        S51_stoich,
        S5_stoich,
        S6,
        S6_stoich,
        S7,
        S7_stoich,
        S8,
        S8_stoich,
        S9,
        S9_stoich,
    ) = variables
    dS0_stoich: float = math.e
    dS1_stoich: float = math.exp(math.e)
    dS2_stoich: float = 1
    dS3_stoich: float = 1
    dS4_stoich: float = math.pi
    dS5_stoich: float = 1.04719755119660
    dS6_stoich: float = (1 / 2) * math.pi
    dS7_stoich: float = -0.523598775598299
    dS8_stoich: float = 1.22777238637419
    dS9_stoich: float = -1.43067687253053
    dS10_stoich: float = 1
    dS11_stoich: float = 4
    dS12_stoich: float = -4
    dS13_stoich: float = -0.947721602131112
    dS14_stoich: float = 0.975897449330606
    dS15_stoich: float = 1
    dS16_stoich: float = math.e
    dS17_stoich: float = 2.15976625378492
    dS18_stoich: float = -5
    dS19_stoich: float = 9
    dS20_stoich: float = -1.60943791243410
    dS21_stoich: float = 0
    dS22_stoich: float = -1.6094379124341 / math.log(10)
    dS23_stoich: float = 0
    dS24_stoich: float = 1
    dS25_stoich: float = 4
    dS26_stoich: float = 5.56228172775440
    dS27_stoich: float = 16
    dS28_stoich: float = 9.61000000000000
    dS29_stoich: float = 2
    dS30_stoich: float = 2.72029410174709
    dS31_stoich: float = 0.863209366648874
    dS32_stoich: float = 0
    dS33_stoich: float = 0.373876664830236
    dS34_stoich: float = 0
    dS35_stoich: float = 2.01433821447683
    dS36_stoich: float = -math.tan(6)
    dS37_stoich: float = 1.13949392732455
    dS38_stoich: float = -1.02298638367130
    dS39_stoich: float = 4.93315487558689
    dS40_stoich: float = 0.304520293447143
    dS41_stoich: float = 2.82831545788997
    dS42_stoich: float = 1.12099946637054
    dS43_stoich: float = 1.14109666064347
    dS44_stoich: float = -1.47112767430373
    dS45_stoich: float = math.asinh(99)
    dS46_stoich: float = 0.802881971289134
    dS47_stoich: float = -0.867300527694053
    dS48_stoich: float = 1.51330668842682
    dS49_stoich: float = 5.29834236561059
    dS50_stoich: float = 1
    dS51_stoich: float = 0
    J0: float = 1
    J1: float = 1
    J2: float = 1
    J3: float = 1
    J4: float = 1
    J5: float = 1
    J6: float = 1
    J7: float = 1
    J8: float = 1
    J9: float = 1
    J10: float = 1
    J11: float = 1
    J12: float = 1
    J13: float = 1
    J14: float = 1
    J15: float = 1
    J16: float = 1
    J17: float = 1
    J18: float = 1
    J19: float = 1
    J20: float = 1
    J21: float = 1
    J22: float = 1
    J23: float = 1
    J24: float = 1
    J25: float = 1
    J26: float = 1
    J27: float = 1
    J28: float = 1
    J29: float = 1
    J30: float = 1
    J31: float = 1
    J32: float = 1
    J33: float = 1
    J34: float = 1
    J35: float = 1
    J36: float = 1
    J37: float = 1
    J38: float = 1
    J39: float = 1
    J40: float = 1
    J41: float = 1
    J42: float = 1
    J43: float = 1
    J44: float = 1
    J45: float = 1
    J46: float = 1
    J47: float = 1
    J48: float = 1
    J49: float = 1
    J50: float = 1
    J51: float = 1
    return {
        "dS0_stoich": dS0_stoich,
        "dS1_stoich": dS1_stoich,
        "dS2_stoich": dS2_stoich,
        "dS3_stoich": dS3_stoich,
        "dS4_stoich": dS4_stoich,
        "dS5_stoich": dS5_stoich,
        "dS6_stoich": dS6_stoich,
        "dS7_stoich": dS7_stoich,
        "dS8_stoich": dS8_stoich,
        "dS9_stoich": dS9_stoich,
        "dS10_stoich": dS10_stoich,
        "dS11_stoich": dS11_stoich,
        "dS12_stoich": dS12_stoich,
        "dS13_stoich": dS13_stoich,
        "dS14_stoich": dS14_stoich,
        "dS15_stoich": dS15_stoich,
        "dS16_stoich": dS16_stoich,
        "dS17_stoich": dS17_stoich,
        "dS18_stoich": dS18_stoich,
        "dS19_stoich": dS19_stoich,
        "dS20_stoich": dS20_stoich,
        "dS21_stoich": dS21_stoich,
        "dS22_stoich": dS22_stoich,
        "dS23_stoich": dS23_stoich,
        "dS24_stoich": dS24_stoich,
        "dS25_stoich": dS25_stoich,
        "dS26_stoich": dS26_stoich,
        "dS27_stoich": dS27_stoich,
        "dS28_stoich": dS28_stoich,
        "dS29_stoich": dS29_stoich,
        "dS30_stoich": dS30_stoich,
        "dS31_stoich": dS31_stoich,
        "dS32_stoich": dS32_stoich,
        "dS33_stoich": dS33_stoich,
        "dS34_stoich": dS34_stoich,
        "dS35_stoich": dS35_stoich,
        "dS36_stoich": dS36_stoich,
        "dS37_stoich": dS37_stoich,
        "dS38_stoich": dS38_stoich,
        "dS39_stoich": dS39_stoich,
        "dS40_stoich": dS40_stoich,
        "dS41_stoich": dS41_stoich,
        "dS42_stoich": dS42_stoich,
        "dS43_stoich": dS43_stoich,
        "dS44_stoich": dS44_stoich,
        "dS45_stoich": dS45_stoich,
        "dS46_stoich": dS46_stoich,
        "dS47_stoich": dS47_stoich,
        "dS48_stoich": dS48_stoich,
        "dS49_stoich": dS49_stoich,
        "dS50_stoich": dS50_stoich,
        "dS51_stoich": dS51_stoich,
        "J0": J0,
        "J1": J1,
        "J2": J2,
        "J3": J3,
        "J4": J4,
        "J5": J5,
        "J6": J6,
        "J7": J7,
        "J8": J8,
        "J9": J9,
        "J10": J10,
        "J11": J11,
        "J12": J12,
        "J13": J13,
        "J14": J14,
        "J15": J15,
        "J16": J16,
        "J17": J17,
        "J18": J18,
        "J19": J19,
        "J20": J20,
        "J21": J21,
        "J22": J22,
        "J23": J23,
        "J24": J24,
        "J25": J25,
        "J26": J26,
        "J27": J27,
        "J28": J28,
        "J29": J29,
        "J30": J30,
        "J31": J31,
        "J32": J32,
        "J33": J33,
        "J34": J34,
        "J35": J35,
        "J36": J36,
        "J37": J37,
        "J38": J38,
        "J39": J39,
        "J40": J40,
        "J41": J41,
        "J42": J42,
        "J43": J43,
        "J44": J44,
        "J45": J45,
        "J46": J46,
        "J47": J47,
        "J48": J48,
        "J49": J49,
        "J50": J50,
        "J51": J51,
    }



time: float = 0.0
x: float = 0.0
y: float = 2.00000000000000

# Initial assignments
z = (2) if (x <= 0.49) else ((1) if (y > 0.49) else (0))
dx = 1
dy = -2
y0 = [x, y]
variable_names = ["x", "y"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    x, y = variables
    dx: float = 1
    dy: float = -2
    dxdt: float = dx
    dydt: float = dy
    return dxdt, dydt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    x, y = variables
    z: float = (2) if (x <= 0.49) else ((1) if (y > 0.49) else (0))
    dx: float = 1
    dy: float = -2
    return {
        "z": z,
        "dx": dx,
        "dy": dy,
    }



time: float = 0.0
C1: float = 0.500000000000000
S1_conc: float = 0.0

# Initial assignments
x = S1_conc
S1 = C1 * S1_conc
dS1 = 0.400000000000000
dC1 = 0.200000000000000
y0 = [C1, S1_conc]
variable_names = ["C1", "S1_conc"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    C1, S1_conc = variables
    dS1: float = 0.400000000000000
    dC1: float = 0.200000000000000
    dS1_concdt: float = dS1
    dC1dt: float = dC1
    return dC1dt, dS1_concdt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    C1, S1_conc = variables
    x: float = S1_conc
    S1: float = C1 * S1_conc
    dS1: float = 0.400000000000000
    dC1: float = 0.200000000000000
    return {
        "x": x,
        "S1": S1,
        "dS1": dS1,
        "dC1": dC1,
    }



time: float = 0.0
k1: float = 0.750000000000000
k2: float = 0.250000000000000
compartment: float = 1.00000000000000
compartment1: float = 1.00000000000000
S1: float = 1.00000000000000
S3: float = 0.0

# Initial assignments
S1_conc = S1 / compartment
S3_conc = S3 / compartment1
reaction1 = S1_conc * k1
reaction2 = k2 * (-S1_conc + S3_conc)
y0 = [S1, S3]
variable_names = ["S1", "S3"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, S3 = variables
    S1_conc: float = S1 / compartment
    S3_conc: float = S3 / compartment1
    reaction1: float = S1_conc * k1
    reaction2: float = k2 * (-S1_conc + S3_conc)
    dS1dt: float = -compartment * reaction1 + compartment * reaction2
    dS3dt: float = compartment1 * reaction1 - compartment1 * reaction2
    return dS1dt, dS3dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, S3 = variables
    S1_conc: float = S1 / compartment
    S3_conc: float = S3 / compartment1
    reaction1: float = S1_conc * k1
    reaction2: float = k2 * (-S1_conc + S3_conc)
    return {
        "S1_conc": S1_conc,
        "S3_conc": S3_conc,
        "reaction1": reaction1,
        "reaction2": reaction2,
    }



time: float = 0.0
k1: float = 1.00000000000000
default_compartment: float = 1.00000000000000

# Initial assignments
J0 = k1
X_amount = 1.0 * default_compartment
p1 = X_amount
X = X_amount / default_compartment
Xref = p1
y0 = [X_amount]
variable_names = ["X_amount"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    (X_amount,) = variables
    p1: float = X_amount
    J0: float = k1
    Xref: float = p1
    dX_amountdt: float = J0 * Xref * default_compartment
    return (dX_amountdt,)


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    (X_amount,) = variables
    p1: float = X_amount
    X: float = X_amount / default_compartment
    J0: float = k1
    Xref: float = p1
    return {
        "p1": p1,
        "X": X,
        "J0": J0,
        "Xref": Xref,
    }



time: float = 0.0
C1: float = 1.50000000000000
C2: float = 1.50000000000000

# Initial assignments
dC1 = -0.2 * C1 + 0.5 * C2
dC2 = 0.75 * C1 - 2.1 * C2
y0 = [C1, C2]
variable_names = ["C1", "C2"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    C1, C2 = variables
    dC1: float = -0.2 * C1 + 0.5 * C2
    dC2: float = 0.75 * C1 - 2.1 * C2
    dC1dt: float = dC1
    dC2dt: float = dC2
    return dC1dt, dC2dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    C1, C2 = variables
    dC1: float = -0.2 * C1 + 0.5 * C2
    dC2: float = 0.75 * C1 - 2.1 * C2
    return {
        "dC1": dC1,
        "dC2": dC2,
    }



time: float = 0.0

# Initial assignments
p1 = max(-5, -time)
p2 = time
p3 = max(5, time)
y0 = []
variable_names = []


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    return


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    p1: float = max(-5, -time)
    p2: float = time
    p3: float = max(5, time)
    return {
        "p1": p1,
        "p2": p2,
        "p3": p3,
    }



time: float = 0.0
kf: float = 2.50000000000000
kr: float = 0.200000000000000
C: float = 1.00000000000000
S1: float = 1.00000000000000
S2: float = 0.500000000000000
S3: float = 0.0

# Initial assignments
S1_conc = S1 / C
S2_conc = S2 / C
S3_conc = S3 / C
reaction1 = S1_conc * kf - S2_conc * S3_conc * kr
y0 = [S2, S3]
variable_names = ["S2", "S3"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S2, S3 = variables
    S1_conc: float = S1 / C
    S2_conc: float = S2 / C
    S3_conc: float = S3 / C
    reaction1: float = S1_conc * kf - S2_conc * S3_conc * kr
    dS2dt: float = C * reaction1
    dS3dt: float = C * reaction1
    return dS2dt, dS3dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S2, S3 = variables
    S1_conc: float = S1 / C
    S2_conc: float = S2 / C
    S3_conc: float = S3 / C
    reaction1: float = S1_conc * kf - S2_conc * S3_conc * kr
    return {
        "S1_conc": S1_conc,
        "S2_conc": S2_conc,
        "S3_conc": S3_conc,
        "reaction1": reaction1,
    }



from __future__ import annotations

import keyword
import re

__all__ = ["RE_FROM_SBML", "RE_KWDS", "SBML_DOT", "name_to_py"]

RE_KWDS = re.compile("|".join(f"^{i}$" for i in keyword.kwlist))
SBML_DOT = "__SBML_DOT__"
RE_FROM_SBML = re.compile(r"__(\d+)__")


def _escape_keyword(re_sub: re.Match) -> str:
    return f"{re_sub.group(0)}_"


def _ascii_to_character(re_sub: re.Match) -> str:
    """Convert an escaped non-alphanumeric character."""
    return chr(int(re_sub.group(1)))


def name_to_py(name: str) -> str:
    name = RE_FROM_SBML.sub(_ascii_to_character, name)
    name = RE_KWDS.sub(_escape_keyword, name)
    name = (
        name.replace(SBML_DOT, ".")
        .replace(" ", "_")
        .replace("-", "_")
        .replace("(", "")
        .replace(")", "")
        .replace("[", "")
        .replace("]", "")
        .replace(".", "")
        .replace(",", "")
        .replace(":", "")
        .replace(";", "")
        .replace('"', "")
        .replace("'", "")
        .replace("^", "")
        .replace("|", "")
        .replace("=", "eq")
        .replace(">", "lg")
        .replace("<", "sm")
        .replace("+", "plus")
        .replace("-", "minus")
        .replace("*", "star")
        .replace("/", "div")
        # .lower()
    )
    if len(name) == 0:
        return name
    if not name[0].isalpha():
        return f"_{name}"
    return name



time: float = 0.0
k1: float = 1.00000000000000
c: float = 1.00000000000000
S1: float = 0.0
S2: float = 0.0

# Initial assignments
J0 = k1
J1 = J0 + 1
y0 = [S1, S2]
variable_names = ["S1", "S2"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, S2 = variables
    J0: float = k1
    J1: float = J0 + 1
    dS1dt: float = J0
    dS2dt: float = J1
    return dS1dt, dS2dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, S2 = variables
    J0: float = k1
    J1: float = J0 + 1
    return {
        "J0": J0,
        "J1": J1,
    }



time: float = 0.0
S3: float = 4.00000000000000
comp: float = 5.00000000000000
S1: float = 1.00000000000000

# Initial assignments
S1_conc = S1 / comp
S3_conc = S3 / comp
dcomp = 1
__J0 = (1 / 10) * S3_conc
y0 = [S1, comp]
variable_names = ["S1", "comp"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, comp = variables
    S3_conc: float = S3 / comp
    dcomp: float = 1
    __J0: float = (1 / 10) * S3_conc
    dcompdt: float = dcomp
    dS1dt: float = -__J0 * comp
    return dS1dt, dcompdt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, comp = variables
    S1_conc: float = S1 / comp
    S3_conc: float = S3 / comp
    dcomp: float = 1
    __J0: float = (1 / 10) * S3_conc
    return {
        "S1_conc": S1_conc,
        "S3_conc": S3_conc,
        "dcomp": dcomp,
        "__J0": __J0,
    }



time: float = 0.0
default_compartment: float = 1.00000000000000
k1: float = 1.00000000000000
Xref: float = 1.00000000000000
X: float = 0.0

# Initial assignments
dXref = 0.0100000000000000
J0 = k1
y0 = [X, Xref]
variable_names = ["X", "Xref"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    X, Xref = variables
    dXref: float = 0.0100000000000000
    J0: float = k1
    dXrefdt: float = dXref
    dXdt: float = J0 * Xref
    return dXdt, dXrefdt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    X, Xref = variables
    dXref: float = 0.0100000000000000
    J0: float = k1
    return {
        "dXref": dXref,
        "J0": J0,
    }



import math

time: float = 0.0

# Initial assignments
P2 = -0.379948962255225
P3 = math.tanh(time)
P4 = -math.tanh(time)
P1 = 0.291312612451591
y0 = []
variable_names = []


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    return


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    P2: float = -0.379948962255225
    P3: float = math.tanh(time)
    P4: float = -math.tanh(time)
    return {
        "P2": P2,
        "P3": P3,
        "P4": P4,
    }



time: float = 0.0
k1: float = 1.00000000000000
default_compartment: float = 1.00000000000000

# Initial assignments
J0 = k1
X_amount = 1.0 * default_compartment
Y_amount = 1.0 * default_compartment
Xref = X_amount
X = X_amount / default_compartment
Y = Y_amount / default_compartment
J1 = Y * k1
y0 = [X_amount, Y_amount]
variable_names = ["X_amount", "Y_amount"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    X_amount, Y_amount = variables
    Xref: float = X_amount
    Y: float = Y_amount / default_compartment
    J0: float = k1
    J1: float = Y * k1
    dX_amountdt: float = J0 * Xref * default_compartment
    dY_amountdt: float = J1 * default_compartment
    return dX_amountdt, dY_amountdt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    X_amount, Y_amount = variables
    Xref: float = X_amount
    X: float = X_amount / default_compartment
    Y: float = Y_amount / default_compartment
    J0: float = k1
    J1: float = Y * k1
    return {
        "Xref": Xref,
        "X": X,
        "Y": Y,
        "J0": J0,
        "J1": J1,
    }



import math

time: float = 0.0
k1: float = 1.00000000000000
k2: float = 0.750000000000000
P1: float = 1.50000000000000
P2: float = 0.0

# Initial assignments
dP1 = P2 * k2 * math.exp(-time)
dP2 = P1 * k1 * math.exp(-time)
y0 = [P1, P2]
variable_names = ["P1", "P2"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    P1, P2 = variables
    dP1: float = P2 * k2 * math.exp(-time)
    dP2: float = P1 * k1 * math.exp(-time)
    dP1dt: float = dP1
    dP2dt: float = dP2
    return dP1dt, dP2dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    P1, P2 = variables
    dP1: float = P2 * k2 * math.exp(-time)
    dP2: float = P1 * k1 * math.exp(-time)
    return {
        "dP1": dP1,
        "dP2": dP2,
    }



"""Manually created model to show best implementation."""

time: float = 0.0
k1: float = 0.750000000000000
k2: float = 0.250000000000000
p1: float = 0.100000000000000
C: float = 1.00000000000000
S1: float = 1.00000000000000
S2: float = 2.00000000000000
S3: float = 1.00000000000000

# Initial assignments
y0 = [C, S1, S2, S3]
variable_names = ["C", "S1", "S2", "S3"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    C, S1, S2, S3 = variables
    dC: float = -C * p1
    reaction1: float = C * S1 * S2 * k1
    reaction2: float = C * S3 * k2
    dCdt: float = dC
    dS1dt: float = -reaction1 / C**2 + reaction2 / C
    dS2dt: float = -reaction1 / C**2 + reaction2 / C
    dS3dt: float = reaction1 / C**2 - reaction2 / C
    return dCdt, dS1dt, dS2dt, dS3dt



time: float = 0.0
C: float = 6.60000000000000
S2: float = 0.150000000000000

# Initial assignments
k1 = 0.750000000000000
S2_conc = S2 / C
S1 = 0.133333333333333 * C * k1
S1_conc = S1 / C
reaction1 = S1_conc * k1
y0 = [S1, S2]
variable_names = ["S1", "S2"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, S2 = variables
    k1: float = 0.750000000000000
    S1_conc: float = S1 / C
    reaction1: float = S1_conc * k1
    dS1dt: float = -C * reaction1
    dS2dt: float = C * reaction1
    return dS1dt, dS2dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, S2 = variables
    k1: float = 0.750000000000000
    S1_conc: float = S1 / C
    S2_conc: float = S2 / C
    reaction1: float = S1_conc * k1
    return {
        "k1": k1,
        "S1_conc": S1_conc,
        "S2_conc": S2_conc,
        "reaction1": reaction1,
    }



time: float = 0.0
S1: float = 1.20000000000000
c: float = 2.00000000000000
S1: float = 1.20000000000000

# Initial assignments
S1_conc = S1 / c
dS1 = 1
dc = 1
y0 = [S1, c]
variable_names = ["S1", "c"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, c = variables
    dS1: float = 1
    dc: float = 1
    dS1dt: float = dS1
    dcdt: float = dc
    return dS1dt, dcdt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, c = variables
    S1_conc: float = S1 / c
    dS1: float = 1
    dc: float = 1
    return {
        "S1_conc": S1_conc,
        "dS1": dS1,
        "dc": dc,
    }



time: float = 0.0
C: float = 1.00000000000000
reaction1_k: float = 1.00000000000000
reaction2_k: float = 2.00000000000000
S1: float = 0.00300000000000000
S2: float = 0.0
S3: float = 0.0

# Initial assignments
reaction1 = S1 * reaction1_k
reaction2 = S2 * reaction2_k
y0 = [S1, S2, S3]
variable_names = ["S1", "S2", "S3"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, S2, S3 = variables
    reaction1: float = S1 * reaction1_k
    reaction2: float = S2 * reaction2_k
    dS1dt: float = -reaction1
    dS2dt: float = reaction1 - reaction2
    dS3dt: float = reaction2
    return dS1dt, dS2dt, dS3dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, S2, S3 = variables
    reaction1: float = S1 * reaction1_k
    reaction2: float = S2 * reaction2_k
    return {
        "reaction1": reaction1,
        "reaction2": reaction2,
    }



time: float = 0.0
k1: float = 750000.000000000
k2: float = 0.250000000000000
C: float = 1.00000000000000
S4: float = 1.00000000000000e-7
S1: float = 1.50000000000000e-6
S2: float = 2.00000000000000e-6
S3: float = 1.50000000000000e-6

# Initial assignments
S1_conc = S1 / C
S2_conc = S2 / C
S3_conc = S3 / C
S4_conc = S4 / C
dS4 = 4.0e-6 * k2
reaction1 = S1_conc * S2_conc * k1
reaction2 = S3_conc * k2
y0 = [S1, S2, S3, S4]
variable_names = ["S1", "S2", "S3", "S4"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, S2, S3, S4 = variables
    S1_conc: float = S1 / C
    S2_conc: float = S2 / C
    S3_conc: float = S3 / C
    dS4: float = 4.0e-6 * k2
    reaction1: float = S1_conc * S2_conc * k1
    reaction2: float = S3_conc * k2
    dS4dt: float = C * dS4
    dS1dt: float = -C * reaction1 + C * reaction2
    dS2dt: float = -2.0 * C * reaction1 + C * reaction2
    dS3dt: float = C * reaction1 - C * reaction2
    return dS1dt, dS2dt, dS3dt, dS4dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, S2, S3, S4 = variables
    S1_conc: float = S1 / C
    S2_conc: float = S2 / C
    S3_conc: float = S3 / C
    S4_conc: float = S4 / C
    dS4: float = 4.0e-6 * k2
    reaction1: float = S1_conc * S2_conc * k1
    reaction2: float = S3_conc * k2
    return {
        "S1_conc": S1_conc,
        "S2_conc": S2_conc,
        "S3_conc": S3_conc,
        "S4_conc": S4_conc,
        "dS4": dS4,
        "reaction1": reaction1,
        "reaction2": reaction2,
    }



time: float = 0.0
k1: float = 1.00000000000000
compartment: float = 1.00000000000000

# Initial assignments
S1_amount = 0.00015 * compartment
S2_amount = 0.0001 * compartment
S1 = S1_amount / compartment
S2 = S2_amount / compartment
reaction1 = S1 * k1
y0 = [S1_amount, S2_amount]
variable_names = ["S1_amount", "S2_amount"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1_amount, S2_amount = variables
    S1: float = S1_amount / compartment
    reaction1: float = S1 * k1
    dS1_amountdt: float = -compartment * reaction1
    dS2_amountdt: float = compartment * reaction1
    return dS1_amountdt, dS2_amountdt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1_amount, S2_amount = variables
    S1: float = S1_amount / compartment
    S2: float = S2_amount / compartment
    reaction1: float = S1 * k1
    return {
        "S1": S1,
        "S2": S2,
        "reaction1": reaction1,
    }



time: float = 0.0
k1: float = 0.693000000000000
k2: float = 0.250000000000000
compartment: float = 1.00000000000000
S3: float = 0.00150000000000000
S4: float = 0.00400000000000000
S1: float = 0.00150000000000000
S2: float = 0.00200000000000000

# Initial assignments
S1_conc = S1 / compartment
S2_conc = S2 / compartment
S3_conc = S3 / compartment
S4_conc = S4 / compartment
dS3 = 0.0005 * k1
dS4 = -0.0005 * k2
reaction1 = S1_conc * k1
y0 = [S1, S2, S3, S4]
variable_names = ["S1", "S2", "S3", "S4"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, S2, S3, S4 = variables
    S1_conc: float = S1 / compartment
    dS3: float = 0.0005 * k1
    dS4: float = -0.0005 * k2
    reaction1: float = S1_conc * k1
    dS3dt: float = compartment * dS3
    dS4dt: float = dS4
    dS1dt: float = -compartment * reaction1
    dS2dt: float = compartment * reaction1
    return dS1dt, dS2dt, dS3dt, dS4dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, S2, S3, S4 = variables
    S1_conc: float = S1 / compartment
    S2_conc: float = S2 / compartment
    S3_conc: float = S3 / compartment
    S4_conc: float = S4 / compartment
    dS3: float = 0.0005 * k1
    dS4: float = -0.0005 * k2
    reaction1: float = S1_conc * k1
    return {
        "S1_conc": S1_conc,
        "S2_conc": S2_conc,
        "S3_conc": S3_conc,
        "S4_conc": S4_conc,
        "dS3": dS3,
        "dS4": dS4,
        "reaction1": reaction1,
    }



time: float = 0.0
k2: float = 0.300000000000000

# Initial assignments
k1 = 4 * k2
y0 = []
variable_names = []


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    return


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    k1: float = 4 * k2
    return {
        "k1": k1,
    }



time: float = 0.0
y: float = 2.00000000000000
x: float = 0.0

# Initial assignments
z = ((2) if (y > 1.49) else (1)) if (x <= 0.49) else (0)
dy = -2
dx = 1
y0 = [x, y]
variable_names = ["x", "y"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    x, y = variables
    dy: float = -2
    dx: float = 1
    dydt: float = dy
    dxdt: float = dx
    return dxdt, dydt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    x, y = variables
    z: float = ((2) if (y > 1.49) else (1)) if (x <= 0.49) else (0)
    dy: float = -2
    dx: float = 1
    return {
        "z": z,
        "dy": dy,
        "dx": dx,
    }



time: float = 0.0

# Initial assignments
J0 = time
p1 = J0 + 1
y0 = []
variable_names = []


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    J0: float = time
    return


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    J0: float = time
    p1: float = J0 + 1
    return {
        "J0": J0,
        "p1": p1,
    }



time: float = 0.0
k1: float = 0.750000000000000
k2: float = 0.250000000000000
C: float = 1.00000000000000
S1_conc: float = 0.0100000000000000
S2_conc: float = 0.0200000000000000
S3: float = 0.0100000000000000

# Initial assignments
S1 = C * S1_conc
S2 = C * S2_conc
reaction1 = S1 * S2 * k1
reaction2 = S3 * k2
y0 = [S3]
variable_names = ["S3"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    (S3,) = variables
    S1: float = C * S1_conc
    S2: float = C * S2_conc
    reaction1: float = S1 * S2 * k1
    reaction2: float = S3 * k2
    dS3dt: float = reaction1 - reaction2
    return (dS3dt,)


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    (S3,) = variables
    S1: float = C * S1_conc
    S2: float = C * S2_conc
    reaction1: float = S1 * S2 * k1
    reaction2: float = S3 * k2
    return {
        "S1": S1,
        "S2": S2,
        "reaction1": reaction1,
        "reaction2": reaction2,
    }



time: float = 0.0
C: float = 2.00000000000000

# Initial assignments
S1_stoich = 0.1 * time
J0 = 0.100000000000000
S1_amount = 2.0 * C
S1 = S1_amount / C
y0 = [S1_amount]
variable_names = ["S1_amount"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    (S1_amount,) = variables
    S1_stoich: float = 0.1 * time
    J0: float = 0.100000000000000
    dS1_amountdt: float = C * J0 * S1_stoich
    return (dS1_amountdt,)


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    (S1_amount,) = variables
    S1_stoich: float = 0.1 * time
    S1: float = S1_amount / C
    J0: float = 0.100000000000000
    return {
        "S1_stoich": S1_stoich,
        "S1": S1,
        "J0": J0,
    }



time: float = 0.0
k1: float = 150000.000000000
k2: float = 50.0000000000000
k3: float = 1.50000000000000
compartment: float = 10.0000000000000

# Initial assignments
S1_amount = 1.0e-5 * compartment
S2_amount = 1.5e-5 * compartment
S3_amount = 1.0e-5 * compartment
S4_amount = 2.25e-5 * compartment
S1 = S1_amount / compartment
S2 = S2_amount / compartment
S3 = S3_amount / compartment
reaction1 = S1 * S2 * k1
reaction2 = S3 * k2
S4 = S2 * k3
y0 = [S1_amount, S2_amount, S3_amount]
variable_names = ["S1_amount", "S2_amount", "S3_amount"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1_amount, S2_amount, S3_amount = variables
    S1: float = S1_amount / compartment
    S2: float = S2_amount / compartment
    S3: float = S3_amount / compartment
    reaction1: float = S1 * S2 * k1
    reaction2: float = S3 * k2
    dS1_amountdt: float = -compartment * reaction1 + compartment * reaction2
    dS2_amountdt: float = -compartment * reaction1 + compartment * reaction2
    dS3_amountdt: float = compartment * reaction1 - compartment * reaction2
    return dS1_amountdt, dS2_amountdt, dS3_amountdt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1_amount, S2_amount, S3_amount = variables
    S1: float = S1_amount / compartment
    S2: float = S2_amount / compartment
    S3: float = S3_amount / compartment
    reaction1: float = S1 * S2 * k1
    reaction2: float = S3 * k2
    S4: float = S2 * k3
    return {
        "S1": S1,
        "S2": S2,
        "S3": S3,
        "reaction1": reaction1,
        "reaction2": reaction2,
        "S4": S4,
    }



time: float = 0.0
kf: float = 0.900000000000000
kr: float = 0.0750000000000000
C: float = 1.00000000000000
S1: float = 1.00000000000000
S2: float = 0.0

# Initial assignments
S1_conc = S1 / C
S2_conc = S2 / C
reaction1 = -S1_conc * kf + S2_conc * kr
y0 = [S2]
variable_names = ["S2"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    (S2,) = variables
    S1_conc: float = S1 / C
    S2_conc: float = S2 / C
    reaction1: float = -S1_conc * kf + S2_conc * kr
    dS2dt: float = -C * reaction1
    return (dS2dt,)


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    (S2,) = variables
    S1_conc: float = S1 / C
    S2_conc: float = S2 / C
    reaction1: float = -S1_conc * kf + S2_conc * kr
    return {
        "S1_conc": S1_conc,
        "S2_conc": S2_conc,
        "reaction1": reaction1,
    }



"""SBML reader for ODE models"""

from __future__ import annotations

from typing import TYPE_CHECKING

from pysbml.parse import load_document

from . import parse, transform

if TYPE_CHECKING:
    from pathlib import Path


def load_model(file: str | Path) -> parse.data.Model:
    return load_document(file).model


def load_and_transform_model(file: str | Path) -> transform.data.Model:
    return transform.transform(load_document(file))



time: float = 0.0
C: float = 1.00000000000000
A: float = 30.0000000000000
B: float = 1.00000000000000

# Initial assignments
J0 = 1
y0 = [A, B]
variable_names = ["A", "B"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    A, B = variables
    J0: float = 1
    dAdt: float = -3.0 * J0
    dBdt: float = 3.0 * J0
    return dAdt, dBdt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    A, B = variables
    J0: float = 1
    return {
        "J0": J0,
    }



time: float = 0.0
C: float = 1.00000000000000
J0_avogadro: float = 0.100000000000000
S1: float = 1.00000000000000

# Initial assignments
J0 = J0_avogadro
y0 = [S1]
variable_names = ["S1"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    (S1,) = variables
    J0: float = J0_avogadro
    dS1dt: float = J0
    return (dS1dt,)


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    (S1,) = variables
    J0: float = J0_avogadro
    return {
        "J0": J0,
    }



time: float = 0.0

# Initial assignments
a = 2
b = 4
c = 1
d = 1
e = 0
y0 = []
variable_names = []


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    return


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    return {}



time: float = 0.0
c: float = 1.00000000000000
p1: float = 0.0
S1: float = 0.0

# Initial assignments
J0 = 1
dp1 = J0
y0 = [S1, p1]
variable_names = ["S1", "p1"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, p1 = variables
    J0: float = 1
    dp1: float = J0
    dp1dt: float = dp1
    dS1dt: float = J0
    return dS1dt, dp1dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, p1 = variables
    J0: float = 1
    dp1: float = J0
    return {
        "J0": J0,
        "dp1": dp1,
    }



import itertools as it
import logging
import math
from collections import defaultdict

import libsbml

import pysbml.parse.data as pdata
from pysbml.parse.data import (
    Assignment,
    AtomicUnit,
    Compartment,
    CompositeUnit,
    Constraint,
    Delay,
    Derived,
    Event,
    Function,
    Model,
    Parameter,
    Priority,
    Reaction,
    Species,
    Trigger,
)
from pysbml.parse.mathml import parse_sbml_math
from pysbml.parse.name_conversion import name_to_py
from pysbml.parse.units import get_unit_conversion

__all__ = [
    "LOGGER",
    "UNIT_CONVERSION",
    "handle_nan",
    "parse",
    "parse_compartments",
    "parse_constraints",
    "parse_events",
    "parse_functions",
    "parse_initial_assignments",
    "parse_parameters",
    "parse_reactions",
    "parse_rules",
    "parse_units",
    "parse_variables",
]

UNIT_CONVERSION = get_unit_conversion()

LOGGER = logging.getLogger(__name__)


def handle_nan(value: float) -> float:
    return math.nan if str(value) == "nan" else value


def parse_constraints(model: Model, lib_model: libsbml.Model) -> None:
    for con in lib_model.getListOfConstraints():
        name = con.getId()

        math, args = None, []
        if (node := con.getMath()) is not None:
            math, args = parse_sbml_math(node)

        message = con.getMessage()
        model.constraints[name] = Constraint(math=math, args=args, message=message)


def _parse_event_trigger(trigger: libsbml.Trigger) -> Trigger | None:
    if trigger is None:
        return None

    math, args = None, []
    if (lib_math := trigger.getMath()) is not None:
        math, args = parse_sbml_math(lib_math)

    return Trigger(
        math=math,
        args=args,
        initial_value=trigger.getInitialValue(),
        persistent=trigger.getPersistent(),
    )


def _parse_event_delay(delay: libsbml.Delay | None) -> Delay | None:
    if delay is None:
        return None

    math, args = None, []
    if (lib_math := delay.getMath()) is not None:
        math, args = parse_sbml_math(lib_math)

    return Delay(math=math, args=args)


def _parse_event_priority(priority: libsbml.Priority) -> Priority | None:
    if priority is None:
        return None

    math, args = None, []
    if (lib_math := priority.getMath()) is not None:
        math, args = parse_sbml_math(lib_math)
    return Priority(math, args)


def _parse_event_assignment(assignment: libsbml.AssignmentRule) -> Assignment:
    variable = assignment.getVariable()
    if (lib_math := assignment.getMath()) is None:
        return Assignment(variable, None, [])

    math, args = parse_sbml_math(lib_math)
    return Assignment(variable, math, args)


def parse_events(model: Model, lib_model: libsbml.Model) -> None:
    for e in lib_model.getListOfEvents():
        name = e.getId()

        model.events[name] = Event(
            trigger=_parse_event_trigger(e.getTrigger()),
            delay=_parse_event_delay(e.getDelay()),
            priority=_parse_event_priority(e.getPriority()),
            assignments=[
                _parse_event_assignment(e.getEventAssignment(i))
                for i in range(e.getNumEventAssignments())
            ],
        )


def parse_units(model: Model, lib_model: libsbml.Model) -> None:
    unit_definition: libsbml.UnitDefinition
    unit: libsbml.Unit

    for unit_definition in lib_model.getListOfUnitDefinitions():
        composite_id = unit_definition.getId()
        local_units = []
        for unit in unit_definition.getListOfUnits():
            atomic_unit = AtomicUnit(
                kind=UNIT_CONVERSION[unit.getKind()],
                scale=unit.getScale(),
                exponent=unit.getExponent(),
                multiplier=unit.getMultiplier(),
            )
            local_units.append(atomic_unit.kind)
            model.atomic_units[atomic_unit.kind] = atomic_unit

        model.composite_units[composite_id] = CompositeUnit(
            sbml_id=composite_id,
            units=local_units,
        )


def parse_compartments(model: Model, lib_model: libsbml.Model) -> None:
    compartment: libsbml.Compartment

    for compartment in lib_model.getListOfCompartments():
        sbml_id = name_to_py(compartment.getId())
        model.compartments[sbml_id] = Compartment(
            name=compartment.getName(),
            dimensions=compartment.getSpatialDimensions(),
            size=handle_nan(compartment.getSize()),
            units=compartment.getUnits(),
            is_constant=compartment.getConstant(),
        )


def parse_parameters(model: Model, lib_model: libsbml.Model) -> None:
    parameter: libsbml.Parameter

    for parameter in lib_model.getListOfParameters():
        model.parameters[name_to_py(parameter.getId())] = Parameter(
            value=handle_nan(parameter.getValue()),
            is_constant=parameter.getConstant(),
            unit=parameter.getUnits(),
        )


def parse_variables(model: Model, lib_model: libsbml.Model) -> None:
    for compound in lib_model.getListOfSpecies():
        compound_id = name_to_py(compound.getId())
        conversion_factor: str | None = (
            cf if bool(cf := compound.getConversionFactor()) else None
        )
        initial_amount = (
            None if str(init := compound.getInitialAmount()) == "nan" else init
        )
        initial_concentration = (
            None if str(init := compound.getInitialConcentration()) == "nan" else init
        )
        has_boundary_condition: bool = compound.getBoundaryCondition()
        if has_boundary_condition:
            model.boundary_species.add(compound_id)

        model.variables[compound_id] = Species(
            compartment=compound.getCompartment(),
            conversion_factor=conversion_factor,
            initial_amount=initial_amount,
            initial_concentration=initial_concentration,
            substance_units=compound.getSubstanceUnits(),
            has_only_substance_units=compound.getHasOnlySubstanceUnits(),
            has_boundary_condition=has_boundary_condition,
            is_constant=compound.getConstant(),
        )


def parse_functions(model: Model, lib_model: libsbml.Model) -> None:
    for func in lib_model.getListOfFunctionDefinitions():
        # Sure, why not just have one name
        name = func.getName()
        sbml_id = func.getId()
        if sbml_id is None or sbml_id == "":
            sbml_id = name
        elif name is None or name == "":
            name = sbml_id
        name = name_to_py(name)

        if (node := func.getMath()) is None:
            msg = f"Function {name} has no math element"
            LOGGER.warning(msg)
            continue

        body, args = parse_sbml_math(node=node)
        model.functions[name] = Function(
            body=body,
            args=args,
        )


def parse_initial_assignments(model: Model, lib_model: libsbml.Model) -> None:
    for assignment in lib_model.getListOfInitialAssignments():
        name = name_to_py(assignment.getSymbol())

        if (node := assignment.getMath()) is None:
            msg = f"Initial assignment {name} has no math element"
            LOGGER.warning(msg)
            continue

        body, args = parse_sbml_math(node)
        model.initial_assignments[name] = Derived(
            body=body,
            args=args,
        )


def _parse_algebraic_rule(model: Model, rule: libsbml.AlgebraicRule) -> None:
    name: str = name_to_py(rule.getId())

    if (node := rule.getMath()) is None:
        msg = f"Algebraic rule {name} has no math element"
        LOGGER.warning(msg)
        return

    body, args = parse_sbml_math(node=node)

    model.algebraic_rules[name] = Derived(
        body=body,
        args=args,
    )


def _parse_assignment_rule(model: Model, rule: libsbml.AssignmentRule) -> None:
    name: str = name_to_py(rule.getId())

    if (node := rule.getMath()) is None:
        msg = f"Assignment rule {name} has no math element"
        LOGGER.warning(msg)
        return

    body, args = parse_sbml_math(node=node)

    model.assignment_rules[name] = Derived(
        body=body,
        args=args,
    )


def _parse_rate_rule(model: Model, rule: libsbml.RateRule) -> None:
    name: str = name_to_py(rule.getId())

    if (node := rule.getMath()) is None:
        msg = f"Rate rule {name} has no math element"
        LOGGER.warning(msg)
        return

    body, args = parse_sbml_math(node=node)

    model.rate_rules[name] = Derived(
        body=body,
        args=args,
    )


def parse_rules(model: Model, sbml_model: libsbml.Model) -> None:
    """Parse rules and separate them by type."""
    for rule in sbml_model.getListOfRules():
        if rule.element_name == "algebraicRule":
            _parse_algebraic_rule(model, rule=rule)
        elif rule.element_name == "assignmentRule":
            _parse_assignment_rule(model, rule=rule)
        elif rule.element_name == "rateRule":
            _parse_rate_rule(model, rule=rule)
        else:
            msg = "Unknown rate type"
            raise ValueError(msg)


def _parse_local_parameters(
    kinetic_law: libsbml.KineticLaw,
) -> dict[str, Parameter]:
    """Parse local parameters."""
    pars = {}

    # Because having one extra container of parameters isn't enough
    for parameter in it.chain(
        kinetic_law.getListOfLocalParameters(),
        kinetic_law.getListOfParameters(),
    ):
        name = name_to_py(parameter.getId())
        pars[name] = Parameter(
            value=handle_nan(parameter.getValue()),
            is_constant=parameter.getConstant(),
            unit=None,
        )
    return pars


def _parse_stoichiometries(
    model: Model, reaction: libsbml.Reaction
) -> dict[str, int | list[tuple[float, str]]]:
    """Parse reaction stoichiometries

    Stoichiometries can be multiple things
    - species
    - boundary species
    - references
    """
    dynamic_stoichiometry: dict[str, list[tuple[float, str]]] = {}
    parsed_reactants: defaultdict[str, int] = defaultdict(int)

    substrate: libsbml.SpeciesReference
    for substrate in reaction.getListOfReactants():
        species = name_to_py(substrate.getSpecies())

        # Only species references have Id set
        if (ref := substrate.getId()) != "":
            model.parameters[ref] = Parameter(
                value=substrate.getStoichiometry(), is_constant=False, unit=None
            )
            # Test 1434 has a case where there are multiple references to the same
            # species...
            dynamic_stoichiometry.setdefault(species, []).append((-1.0, ref))

        # Boundary species can safely be ignored
        elif species not in model.boundary_species:
            factor = substrate.getStoichiometry()
            if str(factor) == "nan":
                msg = f"Cannot parse stoichiometry: {factor} for reaction {reaction.getId()}"
                raise ValueError(msg)
            parsed_reactants[species] -= factor

    product: libsbml.SpeciesReference
    parsed_products: defaultdict[str, int] = defaultdict(int)
    for product in reaction.getListOfProducts():
        species = name_to_py(product.getSpecies())
        if (ref := product.getId()) != "":
            dynamic_stoichiometry.setdefault(species, []).append((1.0, ref))
            model.parameters[ref] = Parameter(
                value=product.getStoichiometry(), is_constant=False, unit=None
            )
        elif species not in model.boundary_species:
            factor = product.getStoichiometry()
            if str(factor) == "nan":
                msg = f"Cannot parse stoichiometry: {factor} for reaction {reaction.getId()}"
                raise ValueError(msg)
            parsed_products[species] += factor

    # Combine stoichiometries
    # Hint: you can't just combine the dictionaries, as you have cases like
    # S1 + S2 -> 2S2, which have to be combined to S1 -> S2
    stoichiometries = dict(parsed_reactants)
    for species, value in parsed_products.items():
        stoichiometries[species] = stoichiometries.get(species, 0) + value

    return stoichiometries | dynamic_stoichiometry


def parse_reactions(model: Model, sbml_model: libsbml.Model) -> None:
    for reaction in sbml_model.getListOfReactions():
        name = name_to_py(reaction.getId())
        kinetic_law = reaction.getKineticLaw()

        if kinetic_law is None:
            msg = f"Reaction {name} has no kinetic law"
            LOGGER.warning(msg)
            continue

        node = kinetic_law.getMath()
        body, args = parse_sbml_math(node=node)

        model.reactions[name] = Reaction(
            body=body,
            args=args,
            stoichiometry=_parse_stoichiometries(model=model, reaction=reaction),
            local_pars=_parse_local_parameters(
                kinetic_law=kinetic_law,
            ),
        )


def parse(
    lib_model: libsbml.Model,
    level: int,  # noqa: ARG001
) -> Model:
    """Parse sbml model."""

    model = pdata.Model(
        name=lib_model.getName(),  # type: ignore
        conversion_factor=None  # type: ignore
        if (conv := lib_model.getConversionFactor()) == ""
        else conv,
    )
    parse_units(model=model, lib_model=lib_model)
    parse_constraints(model=model, lib_model=lib_model)
    parse_events(model=model, lib_model=lib_model)
    parse_compartments(model=model, lib_model=lib_model)
    parse_parameters(model=model, lib_model=lib_model)
    parse_variables(model=model, lib_model=lib_model)
    parse_functions(model=model, lib_model=lib_model)
    parse_initial_assignments(model=model, lib_model=lib_model)
    parse_rules(model=model, sbml_model=lib_model)
    parse_reactions(model=model, sbml_model=lib_model)
    return model



time: float = 0.0
C: float = 1.00000000000000
S1_stoich: float = 2.00000000000000
S1: float = 2.00000000000000
S2: float = 3.00000000000000

# Initial assignments
S1_conc = S1 / C
S2_conc = S2 / C
J0 = 0.0100000000000000
k0 = S1_stoich + 2
y0 = [S1, S2]
variable_names = ["S1", "S2"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    S1, S2 = variables
    J0: float = 0.0100000000000000
    dS2dt: float = C * J0
    dS1dt: float = -C * J0 * S1_stoich
    return dS1dt, dS2dt


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    S1, S2 = variables
    S1_conc: float = S1 / C
    S2_conc: float = S2 / C
    J0: float = 0.0100000000000000
    return {
        "S1_conc": S1_conc,
        "S2_conc": S2_conc,
        "J0": J0,
    }



from __future__ import annotations

from dataclasses import dataclass
from queue import Empty, SimpleQueue
from typing import TYPE_CHECKING, cast

import sympy
from sympy.printing.pycode import pycode

if TYPE_CHECKING:
    from pysbml.transform import data as tdata

INDENT = "    "


@dataclass
class Dependency:
    """Container class for building dependency tree."""

    name: str
    required: set[str]


def _sort_dependencies(
    available: set[str],
    elements: list[Dependency],
) -> list[str]:
    """Sort model elements topologically based on their dependencies.

    Args:
        available: Set of available component names
        elements: List of (name, dependencies, supplier) tuples to sort

    Returns:
        List of element names in dependency order

    Raises:
        SortError: If circular dependencies are detected

    """

    order = []
    # FIXME: what is the worst case here?
    max_iterations = len(elements) ** 2
    queue: SimpleQueue[Dependency] = SimpleQueue()
    for dependency in elements:
        queue.put(dependency)

    last_name = None
    i = 0
    while True:
        try:
            dependency = queue.get_nowait()
        except Empty:
            break
        if dependency.required.issubset(available):
            available.add(dependency.name)
            order.append(dependency.name)

        else:
            if last_name == dependency.name:
                order.append(last_name)
                break
            queue.put(dependency)
            last_name = dependency.name
        i += 1

        # Failure case
        if i > max_iterations:
            unsorted = []
            while True:
                try:
                    unsorted.append(queue.get_nowait().name)
                except Empty:
                    break

            msg = "Fuck"
            raise TypeError(msg)
    return order


def free_symbols(expr: sympy.Expr) -> list[str]:
    return [i.name for i in expr.free_symbols if isinstance(i, sympy.Symbol)]


def codegen_expr(expr: sympy.Expr) -> str:
    return cast(str, pycode(expr, fully_qualified_modules=True)).replace(
        "math.factorial", "scipy.special.factorial"
    )


def codegen_value(val: sympy.Float) -> str:
    return cast(str, pycode(val, fully_qualified_modules=True))


def _to_sympy_types(
    x: str | float | tdata.Expr,
) -> tdata.Expr:
    if isinstance(x, str):
        return sympy.Symbol(x)
    if isinstance(x, float | int):
        return sympy.Float(x)
    return x


def _mul_expr(
    x: str | float | tdata.Expr,
    y: str | float | tdata.Expr,
) -> sympy.Expr:
    return _to_sympy_types(x) * _to_sympy_types(y)  # type: ignore


def codegen(model: tdata.Model) -> str:
    # Do calculations
    variable_names = sorted(model.variables)

    exprs = dict(**model.derived, **{k: rxn.expr for k, rxn in model.reactions.items()})
    init_exprs = dict(exprs, **model.initial_assignments)

    initial_order = _sort_dependencies(
        available=(set(model.parameters) | set(model.variables) | {"time"})
        ^ set(model.initial_assignments),
        elements=[
            Dependency(name=name, required=set(free_symbols(expr)))
            for name, expr in init_exprs.items()
        ],
    )

    order = _sort_dependencies(
        available=set(model.parameters) | set(model.variables) | {"time"},
        elements=[
            Dependency(name=name, required=set(free_symbols(expr)))
            for name, expr in exprs.items()
        ],
    )

    diff_eqs: dict[str, sympy.Expr] = {}
    for name, rxn in model.reactions.items():
        for var, stoich in rxn.stoichiometry.items():
            diff_eqs[var] = sympy.Add(
                diff_eqs.get(var, sympy.Float(0.0)), _mul_expr(stoich, name)
            )

    all_args = set()
    for derived in model.derived.values():
        all_args.update(free_symbols(derived))
    for rxn in model.reactions.values():
        all_args.update(free_symbols(rxn.expr))
        all_args.update(rxn.stoichiometry)
    for diff_eq in diff_eqs.values():
        all_args.update(free_symbols(diff_eq))

    # Not all variables always have an equation, because fuck why
    variable_names = [i for i in variable_names if i in diff_eqs]

    # Actual codegen
    source = [
        "import math",
        "import scipy.special",
        "import pandas as pd",
        "",
        "time: float = 0.0",
    ]

    for name, par in model.parameters.items():
        if name in model.initial_assignments:
            continue
        source.append(f"{name}: float = {codegen_value(par.value)}")

    for name, var in model.variables.items():
        if name in model.initial_assignments:
            continue
        source.append(f"{name}: float = {codegen_value(var.value)}")

    # Initial assignments
    if len(initial_order) > 0:
        source.append("\n# Initial assignments")
    source.extend(
        f"{name} = {codegen_expr(init_exprs[name])}" for name in initial_order
    )

    # Write y0
    source.append(f"y0 = [{', '.join(name for name in variable_names)}]")
    source.append(f"variable_names = {variable_names}")

    # Write main function
    source.append(
        "\ndef model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:",
    )
    if len(variable_names) > 0:
        source.append(
            f"{INDENT}{', '.join(variable_names)} = variables"
            if len(variable_names) > 1
            else f"{INDENT}{variable_names[0]}, = variables"
        )

    for name in order:
        if name not in all_args:
            continue
        source.append(f"{INDENT}{name}: float = {codegen_expr(exprs[name])}")

    for name, eq in diff_eqs.items():
        source.append(
            f"{INDENT}d{name}dt: float = {codegen_expr(cast(sympy.Expr, eq.subs(1.0, 1)))}"
        )

    returns = (f"d{i}dt" for i in variable_names)
    if len(variable_names) == 1:
        source.append(f"{INDENT}return ({next(iter(returns))},)")
    else:
        source.append(f"{INDENT}return {', '.join(returns)}")

    # Additional variables
    source.append(
        "\n\ndef derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:"
    )
    if len(variable_names) > 0:
        source.append(
            f"{INDENT}{', '.join(variable_names)} = variables"
            if len(variable_names) > 1
            else f"{INDENT}{variable_names[0]}, = variables"
        )
    source.extend(
        f"{INDENT}{name}: float = {codegen_expr(exprs[name])}" for name in order
    )

    source.extend(
        [
            f"{INDENT}return {{",
            *(f"{INDENT * 2}{name!r}: {name}," for name in order),
            f"{INDENT}}}",
        ]
    )

    return "\n".join(source)



time: float = 0.0
C1: float = 0.500000000000000
S1_conc: float = 0.0

# Initial assignments
x = S1_conc
S1 = C1 * S1_conc
dS1 = 0.400000000000000
y0 = [S1_conc]
variable_names = ["S1_conc"]


def model(time: float, variables: tuple[float, ...]) -> tuple[float, ...]:
    (S1_conc,) = variables
    dS1: float = 0.400000000000000
    dS1_concdt: float = dS1
    return (dS1_concdt,)


def derived(time: float, variables: tuple[float, ...]) -> dict[str, float]:
    (S1_conc,) = variables
    x: float = S1_conc
    S1: float = C1 * S1_conc
    dS1: float = 0.400000000000000
    return {
        "x": x,
        "S1": S1,
        "dS1": dS1,
    }
